@@ -51,14 +51,17 @@
 #define HAVE_STRTOD 1
 #define HAVE_STRTOL 1
 #define HAVE_STRTOUL 1
+#define HAVE_GETCWD 1
 #define HAVE_FLOCK 1
 #define HAVE_VSNPRINTF 1
+#define HAVE_ISNAN 1
 #define HAVE_FINITE 1
 #define HAVE_FMOD 1
 #define HAVE_FREXP 1
 #define HAVE_HYPOT 1
 #define HAVE_MODF 1
 #define HAVE_WAITPID 1
+#define HAVE_FSYNC 1
 #define HAVE_CHSIZE 1
 #define HAVE_TIMES 1
 #define HAVE__SETJMP 1
