@@ -2,8 +2,8 @@
 
   dir.c -
 
-  $Author: michal $
-  $Date: 2003/01/16 07:34:01 $
+  $Author: dave $
+  $Date: 2003/12/19 15:18:09 $
   created at: Wed Jan  5 09:51:01 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -29,7 +29,12 @@
 # define NAMLEN(dirent) strlen((dirent)->d_name)
 #else
 # define dirent direct
-# define NAMLEN(dirent) (dirent)->d_namlen
+# if !defined __NeXT__
+#  define NAMLEN(dirent) (dirent)->d_namlen
+# else
+#  /* On some versions of NextStep, d_namlen is always zero, so avoid it. */
+#  define NAMLEN(dirent) strlen((dirent)->d_name)
+# endif
 # if HAVE_SYS_NDIR_H
 #  include <sys/ndir.h>
 # endif
@@ -82,20 +87,8 @@ char *strchr _((char*,char));
 
 #if defined DOSISH
 #define isdirsep(c) ((c) == '/' || (c) == '\\')
-static char *
-find_dirsep(s)
-    char *s;
-{
-    while (*s) {
-	if (isdirsep(*s))
-	    return s;
-	s = CharNext(s);
-    }
-    return 0;
-}
 #else
 #define isdirsep(c) ((c) == '/')
-#define find_dirsep(s) strchr(s, '/')
 #endif
 
 static char *
@@ -167,14 +160,14 @@ fnmatch(pat, string, flags)
 		return FNM_NOMATCH;
 
 	    if (!c) {
-		if (pathname && find_dirsep(s))
+		if (pathname && *rb_path_next(s))
 		    return FNM_NOMATCH;
 		else
 		    return 0;
 	    }
 	    else if (ISDIRSEP(c)) {
-		s = find_dirsep(s);
-		if (s) {
+		s = rb_path_next(s);
+		if (*s) {
                     s++;
 		    break;
                 }
@@ -262,6 +255,12 @@ dir_s_alloc(klass)
     return obj;
 }
 
+/*
+ *  call-seq:
+ *     Dir.new( string ) -> aDir
+ *
+ *  Returns a new directory object for the named directory.
+ */
 static VALUE
 dir_initialize(dir, dirname)
     VALUE dir, dirname;
@@ -289,6 +288,17 @@ dir_initialize(dir, dirname)
     return dir;
 }
 
+/*
+ *  call-seq:
+ *     Dir.open( string ) => aDir
+ *     Dir.open( string ) {| aDir | block } => anObject
+ *
+ *  With no block, <code>open</code> is a synonym for
+ *  <code>Dir::new</code>. If a block is present, it is passed
+ *  <i>aDir</i> as a parameter. The directory is closed at the end of
+ *  the block, and <code>Dir::open</code> returns the value of the
+ *  block.
+ */
 static VALUE
 dir_s_open(klass, dirname)
     VALUE klass, dirname;
@@ -315,6 +325,15 @@ dir_closed()
     if (dirp->dir == NULL) dir_closed();\
 } while (0)
 
+/*
+ *  call-seq:
+ *     dir.path => string or nil
+ *
+ *  Returns the path parameter passed to <em>dir</em>'s constructor.
+ *
+ *     d = Dir.new("..")
+ *     d.path   #=> ".."
+ */
 static VALUE
 dir_path(dir)
     VALUE dir;
@@ -326,6 +345,18 @@ dir_path(dir)
     return rb_str_new2(dirp->path);
 }
 
+/*
+ *  call-seq:
+ *     dir.read => string or nil
+ *
+ *  Reads the next entry from <em>dir</em> and returns it as a string.
+ *  Returns <code>nil</code> at the end of the stream.
+ *
+ *     d = Dir.new("testdir")
+ *     d.read   #=> "."
+ *     d.read   #=> ".."
+ *     d.read   #=> "config.h"
+ */
 static VALUE
 dir_read(dir)
     VALUE dir;
@@ -348,6 +379,23 @@ dir_read(dir)
     return Qnil;		/* not reached */
 }
 
+/*
+ *  call-seq:
+ *     dir.each { |filename| block }  => dir
+ *
+ *  Calls the block once for each entry in this directory, passing the
+ *  filename of each entry as a parameter to the block.
+ *
+ *     d = Dir.new("testdir")
+ *     d.each  {|x| puts "Got #{x}" }
+ *
+ *  <em>produces:</em>
+ *
+ *     Got .
+ *     Got ..
+ *     Got config.h
+ *     Got main.rb
+ */
 static VALUE
 dir_each(dir)
     VALUE dir;
@@ -363,6 +411,19 @@ dir_each(dir)
     return dir;
 }
 
+/*
+ *  call-seq:
+ *     dir.pos => integer
+ *     dir.tell => integer
+ *
+ *  Returns the current position in <em>dir</em>. See also
+ *  <code>Dir#seek</code>.
+ *
+ *     d = Dir.new("testdir")
+ *     d.tell   #=> 0
+ *     d.read   #=> "."
+ *     d.tell   #=> 12
+ */
 static VALUE
 dir_tell(dir)
     VALUE dir;
@@ -379,6 +440,20 @@ dir_tell(dir)
 #endif
 }
 
+/*
+ *  call-seq:
+ *     dir.seek( integer ) => dir
+ *
+ *  Seeks to a particular location in <em>dir</em>. <i>integer</i>
+ *  must be a value returned by <code>Dir#tell</code>.
+ *
+ *     d = Dir.new("testdir")   #=> #<Dir:0x401b3c40>
+ *     d.read                   #=> "."
+ *     i = d.tell               #=> 12
+ *     d.read                   #=> ".."
+ *     d.seek(i)                #=> #<Dir:0x401b3c40>
+ *     d.read                   #=> ".."
+ */
 static VALUE
 dir_seek(dir, pos)
     VALUE dir, pos;
@@ -394,6 +469,20 @@ dir_seek(dir, pos)
 #endif
 }
 
+/*
+ *  call-seq:
+ *     dir.pos( integer ) => integer
+ *
+ *  Synonym for <code>Dir#seek</code>, but returns the position
+ *  parameter.
+ *
+ *     d = Dir.new("testdir")   #=> #<Dir:0x401b3c40>
+ *     d.read                   #=> "."
+ *     i = d.pos                #=> 12
+ *     d.read                   #=> ".."
+ *     d.pos = i                #=> 12
+ *     d.read                   #=> ".."
+ */
 static VALUE
 dir_set_pos(dir, pos)
     VALUE dir, pos;
@@ -402,6 +491,17 @@ dir_set_pos(dir, pos)
     return pos;
 }
 
+/*
+ *  call-seq:
+ *     dir.rewind => dir
+ *
+ *  Repositions <em>dir</em> to the first entry.
+ *
+ *     d = Dir.new("testdir")
+ *     d.read     #=> "."
+ *     d.rewind   #=> #<Dir:0x401b3fb0>
+ *     d.read     #=> "."
+ */
 static VALUE
 dir_rewind(dir)
     VALUE dir;
@@ -413,6 +513,16 @@ dir_rewind(dir)
     return dir;
 }
 
+/*
+ *  call-seq:
+ *     dir.close => nil
+ *
+ *  Closes the directory stream. Any further attempts to access
+ *  <em>dir</em> will raise an <code>IOError</code>.
+ *
+ *     d = Dir.new("testdir")
+ *     d.close   #=> nil
+ */
 static VALUE
 dir_close(dir)
     VALUE dir;
@@ -449,6 +559,45 @@ chdir_restore(path)
     return Qnil;
 }
 
+/*
+ *  call-seq:
+ *     Dir.chdir( [ string] ) => 0
+ *     Dir.chdir( [ string] ) {| path | block }  => anObject
+ *
+ *  Changes the current working directory of the process to the given
+ *  string. When called without an argument, changes the directory to
+ *  the value of the environment variable <code>HOME</code>, or
+ *  <code>LOGDIR</code>. <code>SystemCallError</code> (probably
+ *  <code>Errno::ENOENT</code>) if the target directory does not exist.
+ *
+ *  If a block is given, it is passed the name of the new current
+ *  directory, and the block is executed with that as the current
+ *  directory. The original working directory is restored when the block
+ *  exits. The return value of <code>chdir</code> is the value of the
+ *  block. <code>chdir</code> blocks can be nested, but in a
+ *  multi-threaded program an error will be raised if a thread attempts
+ *  to open a <code>chdir</code> block while another thread has one
+ *  open.
+ *
+ *     Dir.chdir("/var/spool/mail")
+ *     puts Dir.pwd
+ *     Dir.chdir("/tmp") do
+ *       puts Dir.pwd
+ *       Dir.chdir("/usr") do
+ *         puts Dir.pwd
+ *       end
+ *       puts Dir.pwd
+ *     end
+ *     puts Dir.pwd
+ *
+ *  <em>produces:</em>
+ *
+ *     /var/spool/mail
+ *     /tmp
+ *     /usr
+ *     /tmp
+ *     /var/spool/mail
+ */
 static VALUE
 dir_s_chdir(argc, argv, obj)
     int argc;
@@ -489,24 +638,62 @@ dir_s_chdir(argc, argv, obj)
     return INT2FIX(0);
 }
 
+/*
+ *  call-seq:
+ *     Dir.getwd => string
+ *     Dir.pwd => string
+ *
+ *  Returns the path to the current working directory of this process as
+ *  a string.
+ *
+ *     Dir.chdir("/tmp")   #=> 0
+ *     Dir.getwd           #=> "/tmp"
+ */
 static VALUE
 dir_s_getwd(dir)
     VALUE dir;
 {
-    char *path = my_getcwd();
-    VALUE cwd = rb_tainted_str_new2(path);
+    char *path;
+    VALUE cwd;
+
+    rb_secure(4);
+    path = my_getcwd();
+    cwd = rb_tainted_str_new2(path);
 
     free(path);
     return cwd;
 }
 
+static void check_dirname _((volatile VALUE *));
+static void
+check_dirname(dir)
+    volatile VALUE *dir;
+{
+    char *path, *pend;
+
+    SafeStringValue(*dir);
+    rb_secure(2);
+    path = RSTRING(*dir)->ptr;
+    if (path && *(pend = rb_path_end(rb_path_skip_prefix(path)))) {
+	*dir = rb_str_new(path, pend - path);
+    }
+}
+
+/*
+ *  call-seq:
+ *     Dir.chroot( string ) => 0
+ *
+ *  Changes this process's idea of the file system root. Only a
+ *  privileged process may make this call. Not available on all
+ *  platforms. On Unix systems, see <code>chroot(2)</code> for more
+ *  information.
+ */
 static VALUE
 dir_s_chroot(dir, path)
     VALUE dir, path;
 {
 #if defined(HAVE_CHROOT) && !defined(__CHECKER__)
-    rb_secure(2);
-    SafeStringValue(path);
+    check_dirname(&path);
 
     if (chroot(RSTRING(path)->ptr) == -1)
 	rb_sys_fail(RSTRING(path)->ptr);
@@ -518,6 +705,19 @@ dir_s_chroot(dir, path)
 #endif
 }
 
+/*
+ *  call-seq:
+ *     Dir.mkdir( string [, integer] ) => 0
+ *
+ *  Makes a new directory named by <i>string</i>, with permissions
+ *  specified by the optional parameter <i>anInteger</i>. The
+ *  permissions may be modified by the value of
+ *  <code>File::umask</code>, and are ignored on NT. Raises a
+ *  <code>SystemCallError</code> if the directory cannot be created. See
+ *  also the discussion of permissions in the class documentation for
+ *  <code>File</code>.
+ *
+ */
 static VALUE
 dir_s_mkdir(argc, argv, obj)
     int argc;
@@ -534,8 +734,7 @@ dir_s_mkdir(argc, argv, obj)
 	mode = 0777;
     }
 
-    SafeStringValue(path);
-    rb_secure(2);
+    check_dirname(&path);
 #ifndef _WIN32
     if (mkdir(RSTRING(path)->ptr, mode) == -1)
 	rb_sys_fail(RSTRING(path)->ptr);
@@ -547,12 +746,20 @@ dir_s_mkdir(argc, argv, obj)
     return INT2FIX(0);
 }
 
+/*
+ *  call-seq:
+ *     Dir.delete( string ) => 0
+ *     Dir.rmdir( string ) => 0
+ *     Dir.unlink( string ) => 0
+ *
+ *  Deletes the named directory. Raises a subclass of
+ *  <code>SystemCallError</code> if the directory isn't empty.
+ */
 static VALUE
 dir_s_rmdir(obj, dir)
     VALUE obj, dir;
 {
-    SafeStringValue(dir);
-    rb_secure(2);
+    check_dirname(&dir);
     if (rmdir(RSTRING(dir)->ptr) < 0)
 	rb_sys_fail(RSTRING(dir)->ptr);
 
@@ -727,7 +934,7 @@ glob_helper(path, sub, flags, func, arg)
 	    struct d_link {
 		char *path;
 		struct d_link *next;
-	    } *tmp, *link = 0;
+	    } *tmp, *link, **tail = &link;
 
 	    base = extract_path(path, p);
 	    if (path == p) dir = ".";
@@ -763,7 +970,7 @@ glob_helper(path, sub, flags, func, arg)
 		free(magic);
 		break;
 	    }
-	    
+
 #if defined DOSISH_DRIVE_LETTER
 #define BASE (*base && !((isdirsep(*base) && !base[1]) || (base[1] == ':' && isdirsep(base[2]) && !base[3])))
 #else
@@ -778,6 +985,7 @@ glob_helper(path, sub, flags, func, arg)
 		    sprintf(buf, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
 		    if (lstat(buf, &st) < 0) {
 			if (errno != ENOENT) rb_sys_warning(buf);
+			free(buf);
 			continue;
 		    }
 		    if (S_ISDIR(st.st_mode)) {
@@ -803,12 +1011,13 @@ glob_helper(path, sub, flags, func, arg)
 		    }
 		    tmp = ALLOC(struct d_link);
 		    tmp->path = buf;
-		    tmp->next = link;
-		    link = tmp;
+		    *tail = tmp;
+		    tail = &tmp->next;
 		}
 	    }
 	    closedir(dirp);
 	  finalize:
+	    *tail = 0;
 	    free(base);
 	    free(magic);
 	    if (link) {
@@ -924,7 +1133,7 @@ push_braces(ary, s, flags)
 	p++;
     }
 
-    if (lbrace) {
+    if (lbrace && rbrace) {
 	int len = strlen(s);
 	buf = xmalloc(len + 1);
 	memcpy(buf, s, lbrace-s);
@@ -999,6 +1208,14 @@ rb_push_glob(str, flags)
     return ary;
 }
 
+/*
+ *  call-seq:
+ *     Dir[ string ] => array
+ *
+ *  Equivalent to calling
+ *  <em>dir</em>.<code>glob(</code><i>string,</i><code>0)</code>.
+ *
+ */
 static VALUE
 dir_s_aref(obj, str)
     VALUE obj, str;
@@ -1006,6 +1223,27 @@ dir_s_aref(obj, str)
     return rb_push_glob(str, 0);
 }
 
+/*
+ *  call-seq:
+ *     Dir.glob( string, [flags] ) => array
+ *     Dir.glob( string, [flags] ) {| filename | block }  => false
+ *
+ *  Returns the filenames found by expanding the pattern given in
+ *  <i>string</i>, either as an <i>array</i> or as parameters to the
+ *  block. Note that this pattern is not a regexp (it's closer to a
+ *  shell glob). See <code>File::fnmatch</code> for
+ *  details of file name matching and the meaning of the <i>flags</i>
+ *  parameter.
+ *
+ *     Dir["config.?"]                     #=> ["config.h"]
+ *     Dir.glob("config.?")                #=> ["config.h"]
+ *     Dir.glob("*.[a-z][a-z]")            #=> ["main.rb"]
+ *     Dir.glob("*.[^r]*")                 #=> ["config.h"]
+ *     Dir.glob("*.{rb,h}")                #=> ["main.rb", "config.h"]
+ *     Dir.glob("*")                       #=> ["config.h", "main.rb"]
+ *     Dir.glob("*", File::FNM_DOTMATCH)   #=> [".", "..", "config.h", "main.rb"]
+ *
+ */
 static VALUE
 dir_s_glob(argc, argv, obj)
     int argc;
@@ -1023,6 +1261,23 @@ dir_s_glob(argc, argv, obj)
     return rb_push_glob(str, flags);
 }
 
+/*
+ *  call-seq:
+ *     Dir.foreach( dirname ) {| filename | block }  => nil
+ *
+ *  Calls the block once for each entry in the named directory, passing
+ *  the filename of each entry as a parameter to the block.
+ *
+ *     Dir.foreach("testdir") {|x| puts "Got #{x}" }
+ *
+ *  <em>produces:</em>
+ *
+ *     Got .
+ *     Got ..
+ *     Got config.h
+ *     Got main.rb
+ *
+ */
 static VALUE
 dir_foreach(io, dirname)
     VALUE io, dirname;
@@ -1034,6 +1289,17 @@ dir_foreach(io, dirname)
     return Qnil;
 }
 
+/*
+ *  call-seq:
+ *     Dir.entries( dirname ) => array
+ *
+ *  Returns an array containing all of the filenames in the given
+ *  directory. Will raise a <code>SystemCallError</code> if the named
+ *  directory doesn't exist.
+ *
+ *     Dir.entries("testdir")   #=> [".", "..", "config.h", "main.rb"]
+ *
+ */
 static VALUE
 dir_entries(io, dirname)
     VALUE io, dirname;
@@ -1044,6 +1310,45 @@ dir_entries(io, dirname)
     return rb_ensure(rb_Array, dir, dir_close, dir);
 }
 
+/*
+ *  call-seq:
+ *     File.fnmatch( pattern, path, [flags] ) => (true or false)
+ *     File.fnmatch?( pattern, path, [flags] ) => (true or false)
+ *
+ *  Returns true if <i>path</i> matches against <i>pattern</i> The
+ *  pattern is not a regular expression; instead it follows rules
+ *  similar to shell filename globbing. It may contain the following
+ *  metacharacters:
+ *
+ *  <i>flags</i> is a bitwise OR of the <code>FNM_xxx</code> parameters.
+ *  The same glob pattern and flags are used by <code>Dir::glob</code>.
+ *
+ *     File.fnmatch('cat',       'cat')        #=> true
+ *     File.fnmatch('cat',       'category')   #=> false
+ *     File.fnmatch('c{at,ub}s', 'cats')       #=> false
+ *     File.fnmatch('c{at,ub}s', 'cubs')       #=> false
+ *     File.fnmatch('c{at,ub}s', 'cat')        #=> false
+ *
+ *     File.fnmatch('c?t',    'cat')                       #=> true
+ *     File.fnmatch('c\?t',   'cat')                       #=> false
+ *     File.fnmatch('c??t',   'cat')                       #=> false
+ *     File.fnmatch('c*',     'cats')                      #=> true
+ *     File.fnmatch('c/ * FIXME * /t', 'c/a/b/c/t')                 #=> true
+ *     File.fnmatch('c*t',    'cat')                       #=> true
+ *     File.fnmatch('c\at',   'cat')                       #=> true
+ *     File.fnmatch('c\at',   'cat', File::FNM_NOESCAPE)   #=> false
+ *     File.fnmatch('a?b',    'a/b')                       #=> true
+ *     File.fnmatch('a?b',    'a/b', File::FNM_PATHNAME)   #=> false
+ *
+ *     File.fnmatch('*',   '.profile')                            #=> false
+ *     File.fnmatch('*',   '.profile', File::FNM_DOTMATCH)        #=> true
+ *     File.fnmatch('*',   'dave/.profile')                       #=> true
+ *     File.fnmatch('*',   'dave/.profile', File::FNM_DOTMATCH)   #=> true
+ *     File.fnmatch('*',   'dave/.profile', File::FNM_PATHNAME)   #=> false
+ *     File.fnmatch('* / FIXME *', 'dave/.profile', File::FNM_PATHNAME)   #=> false
+ *     STRICT = File::FNM_PATHNAME | File::FNM_DOTMATCH
+ *     File.fnmatch('* / FIXME *', 'dave/.profile', STRICT)               #=> true
+ */
 static VALUE
 file_s_fnmatch(argc, argv, obj)
     int argc;
@@ -1068,6 +1373,17 @@ file_s_fnmatch(argc, argv, obj)
     return Qfalse;
 }
 
+/*
+ *  Objects of class <code>Dir</code> are directory streams representing
+ *  directories in the underlying file system. They provide a variety of
+ *  ways to list directories and their contents. See also
+ *  <code>File</code>.
+ *
+ *  The directory used in these examples contains the two regular files
+ *  (<code>config.h</code> and <code>main.rb</code>), the parent
+ *  directory (<code>..</code>), and the directory itself
+ *  (<code>.</code>).
+ */
 void
 Init_Dir()
 {
