@@ -2,8 +2,8 @@
 
   dln.c -
 
-  $Author: michal $
-  $Date: 2003/01/16 07:34:01 $
+  $Author: matz $
+  $Date: 2003/12/20 16:59:09 $
   created at: Tue Jan 18 17:05:06 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -31,7 +31,7 @@ char *dln_argv0;
 #pragma alloca
 #endif
 
-#if defined(HAVE_ALLOCA_H) && !defined(__GNUC__)
+#if defined(HAVE_ALLOCA_H)
 #include <alloca.h>
 #endif
 
@@ -91,7 +91,7 @@ char *getenv();
 
 int eaccess();
 
-#if defined(HAVE_DLOPEN) && !defined(USE_DLN_A_OUT) && !defined(_AIX) && !defined(__APPLE__)
+#if defined(HAVE_DLOPEN) && !defined(USE_DLN_A_OUT) && !defined(_AIX) && !defined(__APPLE__) && !defined(_UNICOSMP)
 /* dynamic load with dlopen() */
 # define USE_DLN_DLOPEN
 #endif
@@ -675,7 +675,7 @@ load_1(fd, disp, need_init)
 	    char *key = sym->n_un.n_name;
 
 	    if (st_lookup(sym_tbl, sym[1].n_un.n_name, &old_sym)) {
-		if (st_delete(undef_tbl, &key, NULL)) {
+		if (st_delete(undef_tbl, (st_data_t*)&key, NULL)) {
 		    unlink_undef(key, old_sym->n_value);
 		    free(key);
 		}
@@ -688,7 +688,7 @@ load_1(fd, disp, need_init)
 		st_foreach(reloc_tbl, reloc_repl, &data);
 
 		st_insert(undef_tbl, strdup(sym[1].n_un.n_name), NULL);
-		if (st_delete(undef_tbl, &key, NULL)) {
+		if (st_delete(undef_tbl, (st_data_t*)&key, NULL)) {
 		    free(key);
 		}
 	    }
@@ -756,7 +756,7 @@ load_1(fd, disp, need_init)
 	    }
 
 	    key = sym->n_un.n_name;
-	    if (st_delete(undef_tbl, &key, NULL) != 0) {
+	    if (st_delete(undef_tbl, (st_data_t*)&key, NULL) != 0) {
 		unlink_undef(key, sym->n_value);
 		free(key);
 	    }
@@ -1139,12 +1139,15 @@ dln_sym(name)
 #include <mach-o/rld.h>
 #else
 #include <mach-o/dyld.h>
+#ifndef NSLINKMODULE_OPTION_BINDNOW
+#define NSLINKMODULE_OPTION_BINDNOW 1
 #endif
 #endif
+#else
 #ifdef __APPLE__
 #include <mach-o/dyld.h>
 #endif
-
+#endif
 
 #if defined _WIN32 && !defined __CYGWIN__
 #include <windows.h>
@@ -1218,7 +1221,7 @@ aix_loaderror(const char *pathname)
     int i,j;
 
     struct errtab { 
-	int errno;
+	int errnum;
 	char *errstr;
     } load_errtab[] = {
 	{L_ERROR_TOOMANY,	"too many errors, rest skipped."},
@@ -1245,7 +1248,7 @@ aix_loaderror(const char *pathname)
     for(i = 0; message[i] && *message[i]; i++) {
 	int nerr = atoi(message[i]);
 	for (j=0; j<LOAD_ERRTAB_LEN; j++) {
-	    if (nerr == load_errtab[i].errno && load_errtab[i].errstr)
+           if (nerr == load_errtab[i].errnum && load_errtab[i].errstr)
 		ERRBUF_APPEND(load_errtab[i].errstr);
 	}
 	while (isdigit(*message[i])) message[i]++; 
@@ -1389,30 +1392,41 @@ dln_load(file)
    Special Thanks...
     Yu tomoak-i@is.aist-nara.ac.jp,
     Mi hisho@tasihara.nest.or.jp,
+    sunshine@sunshineco.com,
     and... Miss ARAI Akino(^^;)
  ----------------------------------------------------*/
 #if defined(NeXT) && (NS_TARGET_MAJOR < 4)/* NeXTSTEP rld functions */
 
     {
+        NXStream* s;
 	unsigned long init_address;
 	char *object_files[2] = {NULL, NULL};
 
 	void (*init_fct)();
 	
-	object_files[0] = file;
+	object_files[0] = (char*)file;
 	
+	s = NXOpenFile(2,NX_WRITEONLY);
+
 	/* Load object file, if return value ==0 ,  load failed*/
-	if(rld_load(NULL, NULL, object_files, NULL) == 0) {
+	if(rld_load(s, NULL, object_files, NULL) == 0) {
+	    NXFlush(s);
+	    NXClose(s);
 	    rb_loaderror("Failed to load %.200s", file);
 	}
 
 	/* lookup the initial function */
-	if(rld_lookup(NULL, buf, &init_address) == 0) {
+	if(rld_lookup(s, buf, &init_address) == 0) {
+	    NXFlush(s);
+	    NXClose(s);
 	    rb_loaderror("Failed to lookup Init function %.200s", file);
 	}
 
-	 /* Cannot call *init_address directory, so copy this value to
-	    funtion pointer */
+	NXFlush(s);
+	NXClose(s);
+
+	/* Cannot call *init_address directory, so copy this value to
+	   funtion pointer */
 	init_fct = (void(*)())init_address;
 	(*init_fct)();
 	return (void*)init_address;
@@ -1591,11 +1605,7 @@ dln_find_exe(fname, path)
     const char *path;
 {
     if (!path) {
-#if defined(__human68k__)
-	path = getenv("path");
-#else
-	path = getenv("PATH");
-#endif
+	path = getenv(PATH_ENV);
     }
 
     if (!path) {
@@ -1666,6 +1676,7 @@ dln_find_1(fname, path, exe_flag)
     const char* mac_fullpath;
 #endif
 
+    if (!fname) return fname;
     if (fname[0] == '/') return fname;
     if (strncmp("./", fname, 2) == 0 || strncmp("../", fname, 3) == 0)
       return fname;
@@ -1741,7 +1752,7 @@ dln_find_1(fname, path, exe_flag)
 	    *bp = '\0';
 	    fprintf(stderr, "\tDirectory \"%s\"\n", fbuf);
 	    fprintf(stderr, "\tFile \"%s\"\n", fname);
-	    continue;
+	    goto next;
 	}
 	memcpy(bp, fname, i + 1);
 
@@ -1795,10 +1806,13 @@ dln_find_1(fname, path, exe_flag)
 #else
 		if (mac_fullpath = _macruby_exist_file_in_libdir_as_posix_name(fbuf))
 		    return mac_fullpath;
+
 #endif
 	    }
 	}
 #endif /* MSDOS or _WIN32 or __human68k__ or __EMX__ */
+
+      next:
 	/* if not, and no other alternatives, life is bleak */
 	if (*ep == '\0') {
 	    return NULL;
