@@ -3,7 +3,7 @@
   bignum.c -
 
   $Author: matz $
-  $Date: 2003/01/31 04:00:14 $
+  $Date: 2004/01/22 18:06:36 $
   created at: Fri Jun 10 00:48:55 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -14,6 +14,9 @@
 
 #include <math.h>
 #include <ctype.h>
+#ifdef HAVE_IEEEFP_H
+#include <ieeefp.h>
+#endif
 
 VALUE rb_cBignum;
 
@@ -33,6 +36,8 @@ VALUE rb_cBignum;
 #define BIGLO(x) ((BDIGIT)((x) & (BIGRAD-1)))
 #define BDIGMAX ((BDIGIT)-1)
 
+#define BIGZEROP(x) (RBIGNUM(x)->len == 0 || (RBIGNUM(x)->len == 1 && BDIGITS(x)[0] == 0))
+
 static VALUE
 bignew_1(klass, len, sign)
     VALUE klass;
@@ -198,12 +203,16 @@ rb_quad_pack(buf, val)
 	long len = RBIGNUM(val)->len;
 	BDIGIT *ds;
 
+	if (len > SIZEOF_LONG_LONG/SIZEOF_BDIGITS) {
+	    len = SIZEOF_LONG/SIZEOF_BDIGITS;
+	}
 	ds = BDIGITS(val);
 	q = 0;
 	while (len--) {
 	    q = BIGUP(q);
 	    q += ds[len];
 	}
+	if (!RBIGNUM(val)->sign) q = -q;
     }
     memcpy(buf, (char*)&q, SIZEOF_LONG_LONG);
 }
@@ -266,7 +275,9 @@ rb_quad_pack(buf, val)
 	val = rb_int2big(FIX2LONG(val));
     }
     len = RBIGNUM(val)->len * SIZEOF_BDIGITS;
-    if (len > QUAD_SIZE) len = QUAD_SIZE;
+    if (len > QUAD_SIZE) {
+	rb_raise(rb_eRangeError, "bignum too big to convert into `quad int'");
+    }
     memcpy(buf, (char*)BDIGITS(val), len);
     if (!RBIGNUM(val)->sign) {
 	len = QUAD_SIZE;
@@ -311,7 +322,8 @@ rb_cstr_to_inum(str, base, badcheck)
 {
     const char *s = str;
     char *end;
-    char sign = 1, c, nondigit = 0;
+    char sign = 1, nondigit = 0;
+    int c;
     BDIGIT_DBL num;
     long len, blen = 1;
     long i;
@@ -373,17 +385,22 @@ rb_cstr_to_inum(str, base, badcheck)
 	    str += 2;
 	}
 	break;
+      case 3:
+	len = 2;
+	break;
       case 8:
-	len = 3;
 	if (str[0] == '0' && (str[1] == 'o'||str[1] == 'O')) {
 	    str += 2;
 	}
+      case 4: case 5: case 6: case 7:
+	len = 3;
 	break;
       case 10:
-	len = 4;
 	if (str[0] == '0' && (str[1] == 'd'||str[1] == 'D')) {
 	    str += 2;
 	}
+      case 9: case 11: case 12: case 13: case 14: case 15:
+	len = 4;
 	break;
       case 16:
 	len = 4;
@@ -391,6 +408,17 @@ rb_cstr_to_inum(str, base, badcheck)
 	    str += 2;
 	}
 	break;
+      default:
+	if (base < 2 || 36 < base) {
+	    rb_raise(rb_eArgError, "illegal radix %d", base);
+	}
+	if (base <= 32) {
+	    len = 5;
+	}
+	else {
+	    len = 6;
+	}
+	break;
     }
     if (*str == '0') {		/* squeeze preceeding 0s */
 	while (*++str == '0');
@@ -405,10 +433,7 @@ rb_cstr_to_inum(str, base, badcheck)
 	if (badcheck) {
 	    if (end == str) goto bad; /* no number */
 	    while (*end && ISSPACE(*end)) end++;
-	    if (*end) {		      /* trailing garbage */
-	      bad:
-		rb_invalid_str(s, "Integer");
-	    }
+	    if (*end) goto bad;	      /* trailing garbage */
 	}
 
 	if (POSFIXABLE(val)) {
@@ -432,42 +457,27 @@ rb_cstr_to_inum(str, base, badcheck)
     zds = BDIGITS(z);
     for (i=len;i--;) zds[i]=0;
     while (c = *str++) {
-	switch (c) {
-	  case '8': case '9':
-	    if (base == 8) {
-		c = base;
-		break;
-	    }
-	  case '0': case '1': case '2': case '3': case '4':
-	  case '5': case '6': case '7': 
-	    c = c - '0';
-	    nondigit = 0;
-	    break;
-	  case 'a': case 'b': case 'c':
-	  case 'd': case 'e': case 'f':
-	    c -= 'a' - 'A';
-	  case 'A': case 'B': case 'C':
-	  case 'D': case 'E': case 'F':
-	    if (base != 16) {
-		nondigit = c;
-		c = base;
-	    }
-	    else {
-		c = c - 'A' + 10;
-		nondigit = 0;
-	    }
-	    break;
-	  case '_':
+	if (c == '_') {
 	    if (badcheck) {
 		if (nondigit) goto bad;
 		nondigit = c;
 	    }
 	    continue;
-	  default:
-	    c = base;
+	}
+	else if (isdigit(c)) {
+	    c -= '0';
+	}
+	else if (islower(c)) {
+	    c -= 'a' - 10;
+	}
+	else if (isupper(c)) {
+	    c -= 'A' - 10;
+	}
+	else {
 	    break;
 	}
 	if (c >= base) break;
+	nondigit = 0;
 	i = 0;
 	num = c;
 	for (;;) {
@@ -487,7 +497,10 @@ rb_cstr_to_inum(str, base, badcheck)
 	str--;
 	if (s+1 < str && str[-1] == '_') goto bad;
 	while (*str && ISSPACE(*str)) str++;
-	if (*str) goto bad;
+	if (*str) {
+	  bad:
+	    rb_invalid_str(s, "Integer");
+	}
     }
 
     return bignorm(z);
@@ -503,7 +516,12 @@ rb_str_to_inum(str, base, badcheck)
     long len;
 
     StringValue(str);
-    s = RSTRING(str)->ptr;
+    if (badcheck) {
+	s = StringValueCStr(str);
+    }
+    else {
+	s = RSTRING(str)->ptr;
+    }
     if (s) {
 	len = RSTRING(str)->len;
 	if (s[len]) {		/* no sentinel somehow */
@@ -513,9 +531,6 @@ rb_str_to_inum(str, base, badcheck)
 	    p[len] = '\0';
 	    s = p;
 	}
-	if (badcheck && len != strlen(s)) {
-	    rb_raise(rb_eArgError, "string for Integer contains null byte");
-	}
     }
     return rb_cstr_to_inum(s, base, badcheck); 
 }
@@ -596,7 +611,7 @@ rb_str2inum(str, base)
     return rb_str_to_inum(str, base, base==0);
 }
 
-static char hexmap[] = "0123456789abcdef";
+const char ruby_digitmap[] = "0123456789abcdefghijklmnopqrstuvwxyz";
 VALUE
 rb_big2str(x, base)
     VALUE x;
@@ -612,28 +627,42 @@ rb_big2str(x, base)
 	return rb_fix2str(x, base);
     }
     i = RBIGNUM(x)->len;
-    if (i == 0 || (i == 1 && BDIGITS(x)[0] == 0)) {
+    if (BIGZEROP(x)) {
 	return rb_str_new2("0");
     }
-    if (base == 10) {
-	j = (SIZEOF_BDIGITS/sizeof(char)*CHAR_BIT*i*241L)/800+2;
-	hbase = 10000;
-    }
-    else if (base == 16) {
-	j = (SIZEOF_BDIGITS/sizeof(char)*CHAR_BIT*i)/4+2;
-	hbase = 0x10000;
-    }
-    else if (base == 8) {
-	j = (SIZEOF_BDIGITS/sizeof(char)*CHAR_BIT*i)+2;
-	hbase = 010000;
-    }
-    else if (base == 2) {
-	j = (SIZEOF_BDIGITS*CHAR_BIT*i)+2;
-	hbase = 020;
-    }
-    else {
+    j = SIZEOF_BDIGITS*CHAR_BIT*i;
+    switch (base) {
+      case 2: break;
+      case 3:
+	j = j * 647L / 1024;
+	break;
+      case 4: case 5: case 6: case 7:
+	j /= 2;
+	break;
+      case 8: case 9:
+	j /= 3;
+	break;
+      case 10: case 11: case 12: case 13: case 14: case 15:
+	j = j * 241L / 800;
+	break;
+      case 16: case 17: case 18: case 19: case 20: case 21:
+      case 22: case 23: case 24: case 25: case 26: case 27:
+      case 28: case 29: case 30: case 31:
+	j /= 4;
+	break;
+      case 32: case 33: case 34: case 35: case 36:
+	j /= 5;
+	break;
+      default:
 	rb_raise(rb_eArgError, "illegal radix %d", base);
+	break;
     }
+    j += 2;
+
+    hbase = base * base;
+#if SIZEOF_BDIGITS > 2
+    hbase *= hbase;
+#endif
 
     t = rb_big_clone(x);
     ds = BDIGITS(t);
@@ -651,10 +680,10 @@ rb_big2str(x, base)
 	    num %= hbase;
 	}
 	if (ds[i-1] == 0) i--;
-	k = 4;
+	k = SIZEOF_BDIGITS;
 	while (k--) {
 	    c = (char)(num % base);
-	    s[--j] = hexmap[(int)c];
+	    s[--j] = ruby_digitmap[(int)c];
 	    num /= base;
 	    if (i == 0 && num == 0) break;
 	}
@@ -667,6 +696,20 @@ rb_big2str(x, base)
     return ss;
 }
 
+/*
+ *  call-seq:
+ *     big.to_s(base=10)   =>  string
+ *  
+ *  Returns a string containing the representation of <i>big</i> radix
+ *  <i>base</i> (2 through 36).
+ *     
+ *     12345654321.to_s         #=> "12345654321"
+ *     12345654321.to_s(2)      #=> "1011011111110110111011110000110001"
+ *     12345654321.to_s(8)      #=> "133766736061"
+ *     12345654321.to_s(16)     #=> "2dfdbbc31"
+ *     78546939656932.to_s(36)  #=> "rubyrules"
+ */
+
 static VALUE
 rb_big_to_s(argc, argv, x)
     int argc;
@@ -683,16 +726,20 @@ rb_big_to_s(argc, argv, x)
 }
 
 static unsigned long
-big2ulong(x, type)
+big2ulong(x, type, check)
     VALUE x;
     char *type;
+    int check;
 {
     long len = RBIGNUM(x)->len;
     BDIGIT_DBL num;
     BDIGIT *ds;
 
-    if (len > SIZEOF_LONG/SIZEOF_BDIGITS)
-	rb_raise(rb_eRangeError, "bignum too big to convert into `%s'", type);
+    if (len > SIZEOF_LONG/SIZEOF_BDIGITS) {
+	if (check)
+	    rb_raise(rb_eRangeError, "bignum too big to convert into `%s'", type);
+	len = SIZEOF_LONG/SIZEOF_BDIGITS;
+    }
     ds = BDIGITS(x);
     num = 0;
     while (len--) {
@@ -702,13 +749,29 @@ big2ulong(x, type)
     return num;
 }
 
+unsigned long
+rb_big2ulong_pack(x)
+    VALUE x;
+{
+    unsigned long num = big2ulong(x, "unsigned long", Qfalse);
+    if (!RBIGNUM(x)->sign) {
+	return -num;
+    }
+    return num;
+}
+
 unsigned long
 rb_big2ulong(x)
     VALUE x;
 {
-    unsigned long num = big2ulong(x, "unsigned long");
+    unsigned long num = big2ulong(x, "unsigned long", Qtrue);
 
-    if (!RBIGNUM(x)->sign) return -num;
+    if (!RBIGNUM(x)->sign) {
+	if ((long)num < 0) {
+	    rb_raise(rb_eRangeError, "bignum out of range of unsigned long");
+	}
+	return -num;
+    }
     return num;
 }
 
@@ -716,10 +779,10 @@ long
 rb_big2long(x)
     VALUE x;
 {
-    unsigned long num = big2ulong(x, "int");
+    unsigned long num = big2ulong(x, "long", Qtrue);
 
     if ((long)num < 0 && (RBIGNUM(x)->sign || (long)num != LONG_MIN)) {
-	rb_raise(rb_eRangeError, "bignum too big to convert into `int'");
+	rb_raise(rb_eRangeError, "bignum too big to convert into `long'");
     }
     if (!RBIGNUM(x)->sign) return -(long)num;
     return num;
@@ -824,11 +887,23 @@ rb_big2dbl(x)
     while (i--) {
 	d = ds[i] + BIGRAD*d;
     }
-    if (isinf(d)) d = HUGE_VAL;
+    if (isinf(d)) {
+	rb_warn("Bignum out of Float range");
+	d = HUGE_VAL;
+    }
     if (!RBIGNUM(x)->sign) d = -d;
     return d;
 }
 
+/*
+ *  call-seq:
+ *     big.to_f -> float
+ *  
+ *  Converts <i>big</i> to a <code>Float</code>. If <i>big</i> doesn't
+ *  fit in a <code>Float</code>, the result is infinity.
+ *     
+ */
+
 static VALUE
 rb_big_to_f(x)
     VALUE x;
@@ -836,6 +911,16 @@ rb_big_to_f(x)
     return rb_float_new(rb_big2dbl(x));
 }
 
+/*
+ *  call-seq:
+ *     big <=> numeric   => -1, 0, +1
+ *  
+ *  Comparison---Returns -1, 0, or +1 depending on whether <i>big</i> is
+ *  less than, equal to, or greater than <i>numeric</i>. This is the
+ *  basis for the tests in <code>Comparable</code>.
+ *     
+ */
+
 static VALUE
 rb_big_cmp(x, y)
     VALUE x, y;
@@ -871,6 +956,17 @@ rb_big_cmp(x, y)
 	    (RBIGNUM(x)->sign ? INT2FIX(-1) : INT2FIX(1));
 }
 
+/*
+ *  call-seq:
+ *     big == obj  => true or false
+ *  
+ *  Returns <code>true</code> only if <i>obj</i> has the same value
+ *  as <i>big</i>. Contrast this with <code>Bignum#eql?</code>, which
+ *  requires <i>obj</i> to be a <code>Bignum</code>.
+ *     
+ *     68719476736 == 68719476736.0   #=> true
+ */
+
 static VALUE
 rb_big_eq(x, y)
     VALUE x, y;
@@ -895,6 +991,17 @@ rb_big_eq(x, y)
     return Qtrue;
 }
 
+/*
+ *  call-seq:
+ *     big.eql?(obj)   => true or false
+ *  
+ *  Returns <code>true</code> only if <i>obj</i> is a
+ *  <code>Bignum</code> with the same value as <i>big</i>. Contrast this
+ *  with <code>Bignum#==</code>, which performs type conversions.
+ *     
+ *     68719476736.eql?(68719476736.0)   #=> false
+ */
+
 static VALUE
 rb_big_eql(x, y)
     VALUE x, y;
@@ -906,6 +1013,13 @@ rb_big_eql(x, y)
     return Qtrue;
 }
 
+/*
+ * call-seq:
+ *    -big   =>  other_big
+ *
+ * Unary minus (returns a new Bignum whose value is 0-big)
+ */
+
 static VALUE
 rb_big_uminus(x)
     VALUE x;
@@ -917,6 +1031,18 @@ rb_big_uminus(x)
     return bignorm(z);
 }
 
+/*
+ * call-seq:
+ *     ~big  =>  integer
+ *
+ * Inverts the bits in big. As Bignums are conceptually infinite
+ * length, the result acts as if it had an infinite number of one
+ * bits to the left. In hex representations, this is displayed
+ * as two periods to the left of the digits.
+ *  
+ *   sprintf("%X", ~0x1122334455)    #=> "..FEEDDCCBBAA"
+ */
+
 static VALUE
 rb_big_neg(x)
     VALUE x;
@@ -1025,6 +1151,13 @@ bigadd(x, y, sign)
     return z;
 }
 
+/*
+ *  call-seq:
+ *     big + other  => Numeric
+ *
+ *  Adds big and other, returning the result.
+ */
+
 VALUE
 rb_big_plus(x, y)
     VALUE x, y;
@@ -1044,6 +1177,13 @@ rb_big_plus(x, y)
     }
 }
 
+/*
+ *  call-seq:
+ *     big - other  => Numeric
+ *
+ *  Subtracts other from big, returning the result.
+ */
+
 VALUE
 rb_big_minus(x, y)
     VALUE x, y;
@@ -1063,6 +1203,13 @@ rb_big_minus(x, y)
     }
 }
 
+/*
+ *  call-seq:
+ *     big * other  => Numeric
+ *
+ *  Multiplies big and other, returning the result.
+ */
+
 VALUE
 rb_big_mul(x, y)
     VALUE x, y;
@@ -1123,8 +1270,8 @@ bigdivrem(x, y, divp, modp)
     BDIGIT_DBL_SIGNED num;
     BDIGIT dd, q;
 
+    if (BIGZEROP(y)) rb_num_zerodiv();
     yds = BDIGITS(y);
-    if (ny == 0 && yds[0] == 0) rb_num_zerodiv();
     if (nx < ny || (nx == ny && BDIGITS(x)[nx - 1] < BDIGITS(y)[ny - 1])) {
 	if (divp) *divp = rb_int2big(0);
 	if (modp) *modp = x;
@@ -1248,8 +1395,7 @@ bigdivmod(x, y, divp, modp)
     VALUE mod;
 
     bigdivrem(x, y, divp, &mod);
-    if (RBIGNUM(x)->sign != RBIGNUM(y)->sign &&
-	!(RBIGNUM(mod)->len == 1 && BDIGITS(mod)[0] == 0)) {
+    if (RBIGNUM(x)->sign != RBIGNUM(y)->sign && !BIGZEROP(mod)) {
 	if (divp) *divp = bigadd(*divp, rb_int2big(1), 0);
 	if (modp) *modp = bigadd(mod, y, 1);
     }
@@ -1259,6 +1405,14 @@ bigdivmod(x, y, divp, modp)
     }
 }
 
+/*
+ *  call-seq:
+ *     big / other     => Numeric
+ *     big.div(other)  => Numeric
+ *
+ *  Divides big by other, returning the result.
+ */
+
 static VALUE
 rb_big_div(x, y)
     VALUE x, y;
@@ -1284,6 +1438,15 @@ rb_big_div(x, y)
     return bignorm(z);
 }
 
+/*
+ *  call-seq:
+ *     big % other         => Numeric
+ *     big.modulo(other)   => Numeric
+ *
+ *  Returns big modulo other. See Numeric.divmod for more
+ *  information.
+ */
+
 static VALUE
 rb_big_modulo(x, y)
     VALUE x, y;
@@ -1306,6 +1469,15 @@ rb_big_modulo(x, y)
     return bignorm(z);
 }
 
+/*
+ *  call-seq:
+ *     big.remainder(numeric)    => number
+ *  
+ *  Returns the remainder after dividing <i>big</i> by <i>numeric</i>.
+ *     
+ *     -1234567890987654321.remainder(13731)      #=> -6966
+ *     -1234567890987654321.remainder(13731.24)   #=> -9906.22531493148
+ */
 static VALUE
 rb_big_remainder(x, y)
     VALUE x, y;
@@ -1328,6 +1500,13 @@ rb_big_remainder(x, y)
     return bignorm(z);
 }
 
+/*
+ *  call-seq:
+ *     big.divmod(numeric)   => array
+ *  
+ *  See <code>Numeric#divmod</code>.
+ *     
+ */
 VALUE
 rb_big_divmod(x, y)
     VALUE x, y;
@@ -1350,6 +1529,18 @@ rb_big_divmod(x, y)
     return rb_assoc_new(bignorm(div), bignorm(mod));
 }
 
+/*
+ *  call-seq:
+ *     big.quo(numeric) -> float
+ *  
+ *  Returns the floating point result of dividing <i>big</i> by
+ *  <i>numeric</i>.
+ *     
+ *     -1234567890987654321.quo(13731)      #=> -89910996357705.5
+ *     -1234567890987654321.quo(13731.24)   #=> -89909424858035.7
+ *     
+ */
+
 static VALUE
 rb_big_quo(x, y)
     VALUE x, y;
@@ -1376,6 +1567,19 @@ rb_big_quo(x, y)
     return rb_float_new(dx / dy);
 }
 
+/*
+ *  call-seq:
+ *     big ** exponent   #=> numeric
+ *
+ *  Raises _big_ to the _exponent_ power (which may be an integer, float,
+ *  or anything that will coerce to a number). The result may be
+ *  a Fixnum, Bignum, or Float
+ *
+ *    123456789 ** 2      #=> 15241578750190521
+ *    123456789 ** 1.2    #=> 5126464716.09932
+ *    123456789 ** -2     #=> 6.5610001194102e-17
+ */
+
 VALUE
 rb_big_pow(x, y)
     VALUE x, y;
@@ -1419,6 +1623,13 @@ rb_big_pow(x, y)
     return rb_float_new(pow(rb_big2dbl(x), d));
 }
 
+/*
+ * call-seq:
+ *     big & numeric   =>  integer
+ *
+ * Performs bitwise +and+ between _big_ and _numeric_.
+ */
+
 VALUE
 rb_big_and(x, y)
     VALUE x, y;
@@ -1428,13 +1639,10 @@ rb_big_and(x, y)
     long i, l1, l2;
     char sign;
 
+    y = rb_to_int(y);
     if (FIXNUM_P(y)) {
 	y = rb_int2big(FIX2LONG(y));
     }
-    else {
-	Check_Type(y, T_BIGNUM);
-    }
-
     if (!RBIGNUM(y)->sign) {
 	y = rb_big_clone(y);
 	get2comp(y, Qtrue);
@@ -1470,6 +1678,13 @@ rb_big_and(x, y)
     return bignorm(z);
 }
 
+/*
+ * call-seq:
+ *     big | numeric   =>  integer
+ *
+ * Performs bitwise +or+ between _big_ and _numeric_.
+ */
+
 VALUE
 rb_big_or(x, y)
     VALUE x, y;
@@ -1479,12 +1694,10 @@ rb_big_or(x, y)
     long i, l1, l2;
     char sign;
 
+    y = rb_to_int(y);
     if (FIXNUM_P(y)) {
 	y = rb_int2big(FIX2LONG(y));
     }
-    else {
-	Check_Type(y, T_BIGNUM);
-    }
 
     if (!RBIGNUM(y)->sign) {
 	y = rb_big_clone(y);
@@ -1522,6 +1735,13 @@ rb_big_or(x, y)
     return bignorm(z);
 }
 
+/*
+ * call-seq:
+ *     big ^ numeric   =>  integer
+ *
+ * Performs bitwise +exclusive or+ between _big_ and _numeric_.
+ */
+
 VALUE
 rb_big_xor(x, y)
     VALUE x, y;
@@ -1531,12 +1751,10 @@ rb_big_xor(x, y)
     long i, l1, l2;
     char sign;
 
+    y = rb_to_int(y);
     if (FIXNUM_P(y)) {
 	y = rb_int2big(FIX2LONG(y));
     }
-    else {
-	Check_Type(y, T_BIGNUM);
-    }
 
     if (!RBIGNUM(y)->sign) {
 	y = rb_big_clone(y);
@@ -1578,6 +1796,13 @@ rb_big_xor(x, y)
 
 static VALUE rb_big_rshift _((VALUE,VALUE));
 
+/*
+ * call-seq:
+ *     big << numeric   =>  integer
+ *
+ * Shifts big left _numeric_ positions (right if _numeric_ is negative).
+ */
+
 VALUE
 rb_big_lshift(x, y)
     VALUE x, y;
@@ -1607,6 +1832,13 @@ rb_big_lshift(x, y)
     return bignorm(z);
 }
 
+/*
+ * call-seq:
+ *     big >> numeric   =>  integer
+ *
+ * Shifts big right _numeric_ positions (left if _numeric_ is negative).
+ */
+
 static VALUE
 rb_big_rshift(x, y)
     VALUE x, y;
@@ -1649,6 +1881,25 @@ rb_big_rshift(x, y)
     return bignorm(z);
 }
 
+/*
+ *  call-seq:
+ *     big[n] -> 0, 1
+ *  
+ *  Bit Reference---Returns the <em>n</em>th bit in the (assumed) binary
+ *  representation of <i>big</i>, where <i>big</i>[0] is the least
+ *  significant bit.
+ *     
+ *     a = 9**15
+ *     50.downto(0) do |n|
+ *       print a[n]
+ *     end
+ *     
+ *  <em>produces:</em>
+ *     
+ *     000101110110100000111000011110010100111100010111001
+ *     
+ */
+
 static VALUE
 rb_big_aref(x, y)
     VALUE x, y;
@@ -1681,6 +1932,13 @@ rb_big_aref(x, y)
     return INT2FIX(0);
 }
 
+/*
+ * call-seq:
+ *   big.hash   => fixnum
+ *
+ * Compute a hash based on the value of _big_.
+ */
+
 static VALUE
 rb_big_hash(x)
     VALUE x;
@@ -1695,6 +1953,10 @@ rb_big_hash(x)
     return LONG2FIX(key);
 }
 
+/*
+ * MISSING: documentation
+ */
+
 static VALUE
 rb_big_coerce(x, y)
     VALUE x, y;
@@ -1710,6 +1972,15 @@ rb_big_coerce(x, y)
     return Qnil;
 }
 
+/*
+ *  call-seq:
+ *     big.abs -> aBignum
+ *  
+ *  Returns the absolute value of <i>big</i>.
+ *     
+ *     -1234567890987654321.abs   #=> 1234567890987654321
+ */
+
 static VALUE
 rb_big_abs(x)
     VALUE x;
@@ -1728,8 +1999,8 @@ rb_big_rand(max, rand_buf)
 {
     VALUE v;
     long len = RBIGNUM(max)->len;
-    
-    if (len == 0 && BDIGITS(max)[0] == 0) {
+
+    if (BIGZEROP(max)) {
 	return rb_float_new(rand_buf[0]);
     }
     v = bignew(len,1);
@@ -1740,6 +2011,18 @@ rb_big_rand(max, rand_buf)
     return rb_big_modulo((VALUE)v, max);
 }
 
+/*
+ *  call-seq:
+ *     big.size -> integer
+ *  
+ *  Returns the number of bytes in the machine representation of
+ *  <i>big</i>.
+ *     
+ *     (256**10 - 1).size   #=> 12
+ *     (256**20 - 1).size   #=> 20
+ *     (256**40 - 1).size   #=> 40
+ */
+
 static VALUE
 rb_big_size(big)
     VALUE big;
@@ -1747,6 +2030,24 @@ rb_big_size(big)
     return LONG2FIX(RBIGNUM(big)->len*SIZEOF_BDIGITS);
 }
 
+/*
+ *  Bignum objects hold integers outside the range of
+ *  Fixnum. Bignum objects are created
+ *  automatically when integer calculations would otherwise overflow a
+ *  Fixnum. When a calculation involving
+ *  Bignum objects returns a result that will fit in a
+ *  Fixnum, the result is automatically converted.
+ *     
+ *  For the purposes of the bitwise operations and <code>[]</code>, a
+ *  Bignum is treated as if it were an infinite-length
+ *  bitstring with 2's complement representation.
+ *     
+ *  While Fixnum values are immediate, Bignum
+ *  objects are not---assignment and parameter passing work with
+ *  references to objects, not the objects themselves.
+ *     
+ */
+
 void
 Init_Bignum()
 {
