@@ -3,7 +3,7 @@
   class.c -
 
   $Author: matz $
-  $Date: 2003/01/31 04:00:14 $
+  $Date: 2004/01/19 09:19:29 $
   created at: Tue Aug 10 15:05:44 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -58,40 +58,43 @@ clone_method(mid, body, tbl)
     return ST_CONTINUE;
 }
 
+/* :nodoc: */
 VALUE
-rb_mod_clone(module)
-    VALUE module;
+rb_mod_init_copy(clone, orig)
+    VALUE clone, orig;
 {
-    NEWOBJ(clone, struct RClass);
-    CLONESETUP(clone, module);
-
-    RCLASS(clone)->super = RCLASS(module)->super;
-    if (RCLASS(module)->iv_tbl) {
+    rb_obj_init_copy(clone, orig);
+    if (!FL_TEST(CLASS_OF(clone), FL_SINGLETON)) {
+	RBASIC(clone)->klass = rb_singleton_class_clone(orig);
+    }
+    RCLASS(clone)->super = RCLASS(orig)->super;
+    if (RCLASS(orig)->iv_tbl) {
 	ID id;
 
-	RCLASS(clone)->iv_tbl = st_copy(RCLASS(module)->iv_tbl);
+	RCLASS(clone)->iv_tbl = st_copy(RCLASS(orig)->iv_tbl);
 	id = rb_intern("__classpath__");
-	st_delete(RCLASS(clone)->iv_tbl, &id, 0);
+	st_delete(RCLASS(clone)->iv_tbl, (st_data_t*)&id, 0);
 	id = rb_intern("__classid__");
-	st_delete(RCLASS(clone)->iv_tbl, &id, 0);
+	st_delete(RCLASS(clone)->iv_tbl, (st_data_t*)&id, 0);
     }
-    if (RCLASS(module)->m_tbl) {
+    if (RCLASS(orig)->m_tbl) {
 	RCLASS(clone)->m_tbl = st_init_numtable();
-	st_foreach(RCLASS(module)->m_tbl, clone_method,
+	st_foreach(RCLASS(orig)->m_tbl, clone_method,
 	  (st_data_t)RCLASS(clone)->m_tbl);
     }
 
-    return (VALUE)clone;
+    return clone;
 }
 
+/* :nodoc: */
 VALUE
-rb_mod_dup(mod)
-    VALUE mod;
+rb_class_init_copy(clone, orig)
+    VALUE clone, orig;
 {
-    VALUE dup = rb_mod_clone(mod);
-
-    RBASIC(dup)->flags = RBASIC(mod)->flags & (T_MASK|FL_TAINT|FL_SINGLETON);
-    return dup;
+    if (RCLASS(clone)->super != 0) {
+	rb_raise(rb_eTypeError, "already initialized class");
+    }
+    return rb_mod_init_copy(clone, orig);
 }
 
 VALUE
@@ -174,12 +177,24 @@ rb_define_class_id(id, super)
 
     if (!super) super = rb_cObject;
     klass = rb_class_new(super);
-    rb_name_class(klass, id);
     rb_make_metaclass(klass, RBASIC(super)->klass);
 
     return klass;
 }
 
+void
+rb_check_inheritable(super)
+    VALUE super;
+{
+    if (TYPE(super) != T_CLASS) {
+	rb_raise(rb_eTypeError, "superclass must be a Class (%s given)",
+		 rb_obj_classname(super));
+    }
+    if (RBASIC(super)->flags & FL_SINGLETON) {
+	rb_raise(rb_eTypeError, "can't make subclass of virtual class");
+    }
+}
+
 VALUE
 rb_class_inherited(super, klass)
     VALUE super, klass;
@@ -197,9 +212,6 @@ rb_define_class(name, super)
     ID id;
 
     id = rb_intern(name);
-    if (rb_autoload_defined(id)) {
-	rb_autoload_load(id);
-    }
     if (rb_const_defined(rb_cObject, id)) {
 	klass = rb_const_get(rb_cObject, id);
 	if (TYPE(klass) != T_CLASS) {
@@ -215,6 +227,8 @@ rb_define_class(name, super)
     }
     klass = rb_define_class_id(id, super);
     st_add_direct(rb_class_tbl, id, klass);
+    rb_name_class(klass, id);
+    rb_const_set(rb_cObject, id, klass);
     rb_class_inherited(super, klass);
 
     return klass;
@@ -231,7 +245,7 @@ rb_define_class_under(outer, name, super)
 
     id = rb_intern(name);
     if (rb_const_defined_at(outer, id)) {
-	klass = rb_const_get(outer, id);
+	klass = rb_const_get_at(outer, id);
 	if (TYPE(klass) != T_CLASS) {
 	    rb_raise(rb_eTypeError, "%s is not a class", name);
 	}
@@ -286,9 +300,6 @@ rb_define_module(name)
     ID id;
 
     id = rb_intern(name);
-    if (rb_autoload_defined(id)) {
-	rb_autoload_load(id);
-    }
     if (rb_const_defined(rb_cObject, id)) {
 	module = rb_const_get(rb_cObject, id);
 	if (TYPE(module) == T_MODULE)
@@ -297,6 +308,7 @@ rb_define_module(name)
     }
     module = rb_define_module_id(id);
     st_add_direct(rb_class_tbl, id, module);
+    rb_const_set(rb_cObject, id, module);
 
     return module;
 }
@@ -311,7 +323,7 @@ rb_define_module_under(outer, name)
 
     id = rb_intern(name);
     if (rb_const_defined_at(outer, id)) {
-	module = rb_const_get(outer, id);
+	module = rb_const_get_at(outer, id);
 	if (TYPE(module) == T_MODULE)
 	    return module;
 	rb_raise(rb_eTypeError, "%s::%s is not a module",
@@ -402,6 +414,23 @@ rb_include_module(klass, module)
     if (changed) rb_clear_cache();
 }
 
+/*
+ *  call-seq:
+ *     mod.included_modules -> array
+ *  
+ *  Returns the list of modules included in <i>mod</i>.
+ *     
+ *     module Mixin
+ *     end
+ *     
+ *     module Outer
+ *       include Mixin
+ *     end
+ *     
+ *     Mixin.included_modules   #=> []
+ *     Outer.included_modules   #=> [Mixin]
+ */
+
 VALUE
 rb_mod_included_modules(mod)
     VALUE mod;
@@ -417,6 +446,25 @@ rb_mod_included_modules(mod)
     return ary;
 }
 
+/*
+ *  call-seq:
+ *     mod.include?(module)    => true or false
+ *  
+ *  Returns <code>true</code> if <i>module</i> is included in
+ *  <i>mod</i> or one of <i>mod</i>'s ancestors.
+ *     
+ *     module A
+ *     end
+ *     class B
+ *       include A
+ *     end
+ *     class C < B
+ *     end
+ *     B.include?(A)   #=> true
+ *     C.include?(A)   #=> true
+ *     A.include?(A)   #=> false
+ */
+
 VALUE
 rb_mod_include_p(mod, mod2)
     VALUE mod;
@@ -433,12 +481,27 @@ rb_mod_include_p(mod, mod2)
     return Qfalse;
 }
 
+/*
+ *  call-seq:
+ *     mod.ancestors -> array
+ *  
+ *  Returns a list of modules included in <i>mod</i> (including
+ *  <i>mod</i> itself).
+ *     
+ *     module Mod
+ *       include Math
+ *       include Comparable
+ *     end
+ *     
+ *     Mod.ancestors    #=> [Mod, Comparable, Math]
+ *     Math.ancestors   #=> [Math]
+ */
+
 VALUE
 rb_mod_ancestors(mod)
     VALUE mod;
 {
-    VALUE ary = rb_ary_new();
-    VALUE p;
+    VALUE p, ary = rb_ary_new();
 
     for (p = mod; p; p = RCLASS(p)->super) {
 	if (FL_TEST(p, FL_SINGLETON))
@@ -453,216 +516,276 @@ rb_mod_ancestors(mod)
     return ary;
 }
 
-#define VISI_CHECK(x,f) (((x)&NOEX_MASK) == (f))
+#define VISI(x) ((x)&NOEX_MASK)
+#define VISI_CHECK(x,f) (VISI(x) == (f))
 
 static int
-ins_methods_i(key, body, ary)
-    ID key;
-    NODE *body;
+ins_methods_push(name, type, ary, visi)
+    ID name;
+    long type;
     VALUE ary;
+    long visi;
 {
-    if (key == ID_ALLOCATOR) return ST_CONTINUE;
-    if (!body->nd_body) {
-	rb_ary_push(ary, Qnil);
-	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
+    if (type == -1) return ST_CONTINUE;
+    switch (visi) {
+      case NOEX_PRIVATE:
+      case NOEX_PROTECTED:
+      case NOEX_PUBLIC:
+	visi = (type == visi);
+	break;
+      default:
+	visi = (type != NOEX_PRIVATE);
+	break;
     }
-    else if (!VISI_CHECK(body->nd_noex, NOEX_PRIVATE)) {
-	VALUE name = rb_str_new2(rb_id2name(key));
-
-	if (!rb_ary_includes(ary, name)) {
-	    rb_ary_push(ary, name);
-	}
-    }
-    else if (nd_type(body->nd_body) == NODE_ZSUPER) {
-	rb_ary_push(ary, Qnil);
-	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
+    if (visi) {
+	rb_ary_push(ary, rb_str_new2(rb_id2name(name)));
     }
     return ST_CONTINUE;
 }
 
 static int
-ins_methods_prot_i(key, body, ary)
-    ID key;
-    NODE *body;
+ins_methods_i(name, type, ary)
+    ID name;
+    long type;
     VALUE ary;
 {
-    if (key == ID_ALLOCATOR) return ST_CONTINUE;
-    if (!body->nd_body) {
-	rb_ary_push(ary, Qnil);
-	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
-    }
-    else if (VISI_CHECK(body->nd_noex, NOEX_PROTECTED)) {
-	VALUE name = rb_str_new2(rb_id2name(key));
-
-	if (!rb_ary_includes(ary, name)) {
-	    rb_ary_push(ary, name);
-	}
-    }
-    else if (nd_type(body->nd_body) == NODE_ZSUPER) {
-	rb_ary_push(ary, Qnil);
-	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
-    }
-    return ST_CONTINUE;
+    return ins_methods_push(name, type, ary, -1); /* everything but private */
 }
 
 static int
-ins_methods_priv_i(key, body, ary)
-    ID key;
-    NODE *body;
+ins_methods_prot_i(name, type, ary)
+    ID name;
+    long type;
     VALUE ary;
 {
-    if (key == ID_ALLOCATOR) return ST_CONTINUE;
-    if (!body->nd_body) {
-	rb_ary_push(ary, Qnil);
-	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
-    }
-    else if (VISI_CHECK(body->nd_noex, NOEX_PRIVATE)) {
-	VALUE name = rb_str_new2(rb_id2name(key));
-
-	if (!rb_ary_includes(ary, name)) {
-	    rb_ary_push(ary, name);
-	}
-    }
-    else if (nd_type(body->nd_body) == NODE_ZSUPER) {
-	rb_ary_push(ary, Qnil);
-	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
-    }
-    return ST_CONTINUE;
+    return ins_methods_push(name, type, ary, NOEX_PROTECTED);
 }
 
 static int
-ins_methods_pub_i(key, body, ary)
-    ID key;
-    NODE *body;
+ins_methods_priv_i(name, type, ary)
+    ID name;
+    long type;
     VALUE ary;
 {
-    if (key == ID_ALLOCATOR) return ST_CONTINUE;
-    if (!body->nd_body) {
-	rb_ary_push(ary, Qnil);
-	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
-    }
-    else if (VISI_CHECK(body->nd_noex, NOEX_PUBLIC)) {
-	VALUE name = rb_str_new2(rb_id2name(key));
+    return ins_methods_push(name, type, ary, NOEX_PRIVATE);
+}
 
-	if (!rb_ary_includes(ary, name)) {
-	    rb_ary_push(ary, name);
-	}
-    }
-    else if (nd_type(body->nd_body) == NODE_ZSUPER) {
-	rb_ary_push(ary, Qnil);
-	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
+static int
+ins_methods_pub_i(name, type, ary)
+    ID name;
+    long type;
+    VALUE ary;
+{
+    return ins_methods_push(name, type, ary, NOEX_PUBLIC);
+}
+
+static int
+method_entry(key, body, list)
+    ID key;
+    NODE *body;
+    st_table *list;
+{
+    long type;
+
+    if (key == ID_ALLOCATOR) return ST_CONTINUE;
+    if (!st_lookup(list, key, 0)) {
+	if (!body->nd_body) type = -1; /* none */
+	else type = VISI(body->nd_noex);
+	st_add_direct(list, key, type);
     }
     return ST_CONTINUE;
 }
 
 static VALUE
-method_list(mod, option, func)
+class_instance_method_list(argc, argv, mod, func)
+    int argc;
+    VALUE *argv;
     VALUE mod;
-    int option;
-    int (*func)();
+    int (*func) _((ID, long, VALUE));
 {
     VALUE ary;
-    VALUE klass;
-    VALUE *p, *q, *pend;
+    int recur;
+    st_table *list;
 
+    if (argc == 0) {
+	recur = Qtrue;
+    }
+    else {
+	VALUE r;
+	rb_scan_args(argc, argv, "01", &r);
+	recur = RTEST(r);
+    }
+
+    list = st_init_numtable();
+    for (; mod; mod = RCLASS(mod)->super) {
+	st_foreach(RCLASS(mod)->m_tbl, method_entry, (st_data_t)list);
+	if (BUILTIN_TYPE(mod) == T_ICLASS) continue;
+	if (FL_TEST(mod, FL_SINGLETON)) continue;
+	if (!recur) break;
+    }
     ary = rb_ary_new();
-    for (klass = mod; klass; klass = RCLASS(klass)->super) {
-	st_foreach(RCLASS(klass)->m_tbl, func, ary);
-	if (!option) break;
-    }
-    p = q = RARRAY(ary)->ptr; pend = p + RARRAY(ary)->len;
-    while (p < pend) {
-	if (*p == Qnil) {
-	    p+=2;
-	    continue;
-	}
-	*q++ = *p++;
-    }
-    RARRAY(ary)->len = q - RARRAY(ary)->ptr;
+    st_foreach(list, func, ary);
+    st_free_table(list);
+
     return ary;
 }
 
+/*
+ *  call-seq:
+ *     mod.instance_methods(include_super=false)   => array
+ *  
+ *  Returns an array containing the names of public instance methods in
+ *  the receiver. For a module, these are the public methods; for a
+ *  class, they are the instance (not singleton) methods. With no
+ *  argument, or with an argument that is <code>false</code>, the
+ *  instance methods in <i>mod</i> are returned, otherwise the methods
+ *  in <i>mod</i> and <i>mod</i>'s superclasses are returned.
+ *     
+ *     module A
+ *       def method1()  end
+ *     end
+ *     class B
+ *       def method2()  end
+ *     end
+ *     class C < B
+ *       def method3()  end
+ *     end
+ *     
+ *     A.instance_methods                #=> ["method1"]
+ *     B.instance_methods                #=> ["method2"]
+ *     C.instance_methods                #=> ["method3"]
+ *     C.instance_methods(true).length   #=> 43
+ */
+
 VALUE
 rb_class_instance_methods(argc, argv, mod)
     int argc;
     VALUE *argv;
     VALUE mod;
 {
-    VALUE option;
-
-    rb_scan_args(argc, argv, "01", &option);
-    return method_list(mod, RTEST(option), ins_methods_i);
+    return class_instance_method_list(argc, argv, mod, ins_methods_i);
 }
 
+/*
+ *  call-seq:
+ *     mod.protected_instance_methods(include_super=false)   => array
+ *  
+ *  Returns a list of the protected instance methods defined in
+ *  <i>mod</i>. If the optional parameter is not <code>false</code>, the
+ *  methods of any ancestors are included.
+ */
+
 VALUE
 rb_class_protected_instance_methods(argc, argv, mod)
     int argc;
     VALUE *argv;
     VALUE mod;
 {
-    VALUE option;
-
-    rb_scan_args(argc, argv, "01", &option);
-    return method_list(mod, RTEST(option), ins_methods_prot_i);
+    return class_instance_method_list(argc, argv, mod, ins_methods_prot_i);
 }
 
+/*
+ *  call-seq:
+ *     mod.private_instance_methods(include_super=false)    => array
+ *  
+ *  Returns a list of the private instance methods defined in
+ *  <i>mod</i>. If the optional parameter is not <code>false</code>, the
+ *  methods of any ancestors are included.
+ *     
+ *     module Mod
+ *       def method1()  end
+ *       private :method1
+ *       def method2()  end
+ *     end
+ *     Mod.instance_methods           #=> ["method2"]
+ *     Mod.private_instance_methods   #=> ["method1"]
+ */
+
 VALUE
 rb_class_private_instance_methods(argc, argv, mod)
     int argc;
     VALUE *argv;
     VALUE mod;
 {
-    VALUE option;
-
-    rb_scan_args(argc, argv, "01", &option);
-    return method_list(mod, RTEST(option), ins_methods_priv_i);
+    return class_instance_method_list(argc, argv, mod, ins_methods_priv_i);
 }
 
+/*
+ *  call-seq:
+ *     mod.public_instance_methods(include_super=false)   => array
+ *  
+ *  Returns a list of the public instance methods defined in <i>mod</i>.
+ *  If the optional parameter is not <code>false</code>, the methods of
+ *  any ancestors are included.
+ */
+
 VALUE
 rb_class_public_instance_methods(argc, argv, mod)
     int argc;
     VALUE *argv;
     VALUE mod;
 {
-    VALUE option;
-
-    rb_scan_args(argc, argv, "01", &option);
-    return method_list(mod, RTEST(option), ins_methods_pub_i);
+    return class_instance_method_list(argc, argv, mod, ins_methods_pub_i);
 }
 
+/*
+ *  call-seq:
+ *     obj.singleton_methods(all=false)    => array
+ *  
+ *  Returns an array of the names of singleton methods for <i>obj</i>.
+ *  If the optional <i>all</i> parameter is true, the list will include
+ *  methods in modules included in <i>obj</i>.
+ *     
+ *     module Other
+ *       def three() end
+ *     end
+ *     
+ *     class Single
+ *       def Single.four() end
+ *     end
+ *     
+ *     a = Single.new
+ *     
+ *     def a.one()
+ *     
+ *     class << a
+ *       include Other
+ *       def two()
+ *     end
+ *     
+ *     Single.singleton_methods    #=> ["four"]
+ *     a.singleton_methods         #=> ["two", "one"]
+ *     a.singleton_methods(true)   #=> ["two", "one", "three"]
+ */
+
 VALUE
 rb_obj_singleton_methods(argc, argv, obj)
     int argc;
     VALUE *argv;
     VALUE obj;
 {
-    VALUE all;
-    VALUE ary;
-    VALUE klass;
-    VALUE *p, *q, *pend;
+    VALUE recur, ary, klass;
+    st_table *list;
 
-    rb_scan_args(argc, argv, "01", &all);
-    ary = rb_ary_new();
+    rb_scan_args(argc, argv, "01", &recur);
+    if (argc == 0) {
+	recur = Qtrue;
+    }
     klass = CLASS_OF(obj);
-    while (klass && FL_TEST(klass, FL_SINGLETON)) {
-	st_foreach(RCLASS(klass)->m_tbl, ins_methods_i, ary);
+    list = st_init_numtable();
+    if (klass && FL_TEST(klass, FL_SINGLETON)) {
+	st_foreach(RCLASS(klass)->m_tbl, method_entry, (st_data_t)list);
 	klass = RCLASS(klass)->super;
     }
-    if (RTEST(all)) {
-	while (klass && TYPE(klass) == T_ICLASS) {
-	    st_foreach(RCLASS(klass)->m_tbl, ins_methods_i, ary);
+    if (RTEST(recur)) {
+	while (klass && (FL_TEST(klass, FL_SINGLETON) || TYPE(klass) == T_ICLASS)) {
+	    st_foreach(RCLASS(klass)->m_tbl, method_entry, (st_data_t)list);
 	    klass = RCLASS(klass)->super;
 	}
     }
-    p = q = RARRAY(ary)->ptr; pend = p + RARRAY(ary)->len;
-    while (p < pend) {
-	if (*p == Qnil) {
-	    p+=2;
-	    continue;
-	}
-	*q++ = *p++;
-    }
-    RARRAY(ary)->len = q - RARRAY(ary)->ptr;
+    ary = rb_ary_new();
+    st_foreach(list, ins_methods_i, ary);
+    st_free_table(list);
 
     return ary;
 }
@@ -674,7 +797,7 @@ rb_define_method_id(klass, name, func, argc)
     VALUE (*func)();
     int argc;
 {
-    rb_add_method(klass, name, NEW_CFUNC(func,argc), NOEX_PUBLIC|NOEX_CFUNC);
+    rb_add_method(klass, name, NEW_CFUNC(func,argc), NOEX_PUBLIC);
 }
 
 void
@@ -688,9 +811,6 @@ rb_define_method(klass, name, func, argc)
     int ex = NOEX_PUBLIC;
 
 
-    if (BUILTIN_TYPE(klass) == T_CLASS) {
-	ex |= NOEX_CFUNC;
-    }
     rb_add_method(klass, id, NEW_CFUNC(func, argc), ex);
 }
 
@@ -701,8 +821,7 @@ rb_define_protected_method(klass, name, func, argc)
     VALUE (*func)();
     int argc;
 {
-    rb_add_method(klass, rb_intern(name), NEW_CFUNC(func, argc),
-		  NOEX_PROTECTED|NOEX_CFUNC);
+    rb_add_method(klass, rb_intern(name), NEW_CFUNC(func, argc), NOEX_PROTECTED);
 }
 
 void
@@ -712,8 +831,7 @@ rb_define_private_method(klass, name, func, argc)
     VALUE (*func)();
     int argc;
 {
-    rb_add_method(klass, rb_intern(name), NEW_CFUNC(func, argc),
-		  NOEX_PRIVATE|NOEX_CFUNC);
+    rb_add_method(klass, rb_intern(name), NEW_CFUNC(func, argc), NOEX_PRIVATE);
 }
 
 void
@@ -885,7 +1003,7 @@ rb_scan_args(argc, argv, fmt, va_alist)
     if (*p == '&') {
 	var = va_arg(vargs, VALUE*);
 	if (rb_block_given_p()) {
-	    *var = rb_f_lambda();
+	    *var = rb_block_proc();
 	}
 	else {
 	    *var = Qnil;
