@@ -3,7 +3,7 @@
   gc.c -
 
   $Author: matz $
-  $Date: 2003/02/27 08:04:32 $
+  $Date: 2004/01/22 08:31:32 $
   created at: Tue Oct  5 09:44:46 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -30,6 +30,20 @@
 #include <sys/resource.h>
 #endif
 
+#ifdef __ia64__
+#include <ucontext.h>
+#if defined(__FreeBSD__)
+/*
+ * FreeBSD/ia64 currently does not have a way for a process to get the
+ * base address for the RSE backing store, so hardcode it.
+ */
+#define __libc_ia64_register_backing_store_base (4ULL<<61)
+#else
+#pragma weak __libc_ia64_register_backing_store_base
+extern unsigned long __libc_ia64_register_backing_store_base;
+#endif
+#endif
+
 void re_free_registers _((struct re_registers*));
 void rb_io_fptr_finalize _((struct OpenFile*));
 
@@ -162,6 +176,20 @@ static int during_gc;
 static int need_call_final = 0;
 static st_table *finalizer_table = 0;
 
+
+/*
+ *  call-seq:
+ *     GC.enable    => true or false
+ *  
+ *  Enables garbage collection, returning <code>true</code> if garbage
+ *  collection was previously disabled.
+ *     
+ *     GC.disable   #=> false
+ *     GC.enable    #=> true
+ *     GC.enable    #=> false
+ *     
+ */
+
 VALUE
 rb_gc_enable()
 {
@@ -171,6 +199,18 @@ rb_gc_enable()
     return old;
 }
 
+/*
+ *  call-seq:
+ *     GC.disable    => true or false
+ *  
+ *  Disables garbage collection, returning <code>true</code> if garbage
+ *  collection was already disabled.
+ *     
+ *     GC.disable   #=> false
+ *     GC.disable   #=> true
+ *     
+ */
+
 VALUE
 rb_gc_disable()
 {
@@ -222,6 +262,8 @@ rb_gc_unregister_address(addr)
     }
 }
 
+#undef GC_DEBUG
+
 void
 rb_global_variable(var)
     VALUE *var;
@@ -252,18 +294,24 @@ typedef struct RVALUE {
 	struct RVarmap varmap; 
 	struct SCOPE   scope;
     } as;
+#ifdef GC_DEBUG
+    char *file;
+    int   line;
+#endif
 } RVALUE;
 
 static RVALUE *freelist = 0;
 static RVALUE *deferred_final_list = 0;
 
 #define HEAPS_INCREMENT 10
-static RVALUE **heaps;
+static struct heaps_slot {
+    RVALUE *slot;
+    int limit;
+} *heaps;
 static int heaps_length = 0;
 static int heaps_used   = 0;
 
 #define HEAP_MIN_SLOTS 10000
-static int *heaps_limits;
 static int heap_slots = HEAP_MIN_SLOTS;
 
 #define FREE_MIN  4096
@@ -277,20 +325,25 @@ add_heap()
 
     if (heaps_used == heaps_length) {
 	/* Realloc heaps */
+	struct heaps_slot *p;
+	int length;
+
 	heaps_length += HEAPS_INCREMENT;
-	RUBY_CRITICAL(heaps = (heaps_used>0)?
-			(RVALUE**)realloc(heaps, heaps_length*sizeof(RVALUE*)):
-			(RVALUE**)malloc(heaps_length*sizeof(RVALUE*)));
-	if (heaps == 0) rb_memerror();
-	RUBY_CRITICAL(heaps_limits = (heaps_used>0)?
-			(int*)realloc(heaps_limits, heaps_length*sizeof(int)):
-			(int*)malloc(heaps_length*sizeof(int)));
-	if (heaps_limits == 0) rb_memerror();
+	length = heaps_length*sizeof(struct heaps_slot);
+	RUBY_CRITICAL(
+	    if (heaps_used > 0) {
+		p = (struct heaps_slot *)realloc(heaps, length);
+		if (p) heaps = p;
+	    }
+	    else {
+		p = heaps = (struct heaps_slot *)malloc(length);
+	    });
+	if (p == 0) rb_memerror();
     }
 
     for (;;) {
-	RUBY_CRITICAL(p = heaps[heaps_used] = (RVALUE*)malloc(sizeof(RVALUE)*heap_slots));
-	heaps_limits[heaps_used] = heap_slots;
+	RUBY_CRITICAL(p = heaps[heaps_used].slot = (RVALUE*)malloc(sizeof(RVALUE)*heap_slots));
+	heaps[heaps_used].limit = heap_slots;
 	if (p == 0) {
 	    if (heap_slots == HEAP_MIN_SLOTS) {
 		rb_memerror();
@@ -325,6 +378,10 @@ rb_newobj()
     obj = (VALUE)freelist;
     freelist = freelist->as.free.next;
     MEMZERO((void*)obj, RVALUE, 1);
+#ifdef GC_DEBUG
+    RANY(obj)->file = ruby_sourcefile;
+    RANY(obj)->line = ruby_sourceline;
+#endif
     return obj;
 }
 
@@ -336,6 +393,7 @@ rb_data_object_alloc(klass, datap, dmark, dfree)
     RUBY_DATA_FUNC dfree;
 {
     NEWOBJ(data, struct RData);
+    Check_Type(klass, T_CLASS);
     OBJSETUP(data, klass, T_DATA);
     data->data = datap;
     data->dfree = dfree;
@@ -347,7 +405,7 @@ rb_data_object_alloc(klass, datap, dmark, dfree)
 extern st_table *rb_class_tbl;
 VALUE *rb_gc_stack_start = 0;
 
-#ifdef DJGPP
+#if defined(DJGPP) || defined(_WIN32_WCE)
 static unsigned int STACK_LEVEL_MAX = 65535;
 #else
 #ifdef __human68k__
@@ -367,7 +425,7 @@ static unsigned int STACK_LEVEL_MAX = 655300;
 # define SET_STACK_END VALUE stack_end; alloca(0);
 # define STACK_END (&stack_end)
 #else
-# if defined(__GNUC__) && defined(USE_BUILTIN_FRAME_ADDRESS)
+# if defined(__GNUC__) && defined(USE_BUILTIN_FRAME_ADDRESS) && !defined(__ia64__)
 #  define  SET_STACK_END    VALUE *stack_end = __builtin_frame_address(0)
 # else
 #  define  SET_STACK_END    VALUE *stack_end = alloca(1)
@@ -376,14 +434,36 @@ static unsigned int STACK_LEVEL_MAX = 655300;
 #endif
 #if defined(sparc) || defined(__sparc__)
 # define STACK_LENGTH  (rb_gc_stack_start - STACK_END + 0x80)
+#elif STACK_GROW_DIRECTION < 0
+# define STACK_LENGTH  (rb_gc_stack_start - STACK_END)
+#elif STACK_GROW_DIRECTION > 0
+# define STACK_LENGTH  (STACK_END - rb_gc_stack_start + 1)
 #else
 # define STACK_LENGTH  ((STACK_END < rb_gc_stack_start) ? rb_gc_stack_start - STACK_END\
-                                           : STACK_END - rb_gc_stack_start)
+                                           : STACK_END - rb_gc_stack_start + 1)
 #endif
+#if STACK_GROW_DIRECTION > 0
+# define STACK_UPPER(x, a, b) a
+#elif STACK_GROW_DIRECTION < 0
+# define STACK_UPPER(x, a, b) b
+#else
+static int
+stack_growup_p(addr)
+    VALUE *addr;
+{
+    SET_STACK_END;
+
+    if (STACK_END > addr) return Qtrue;
+    return Qfalse;
+}
+# define STACK_UPPER(x, a, b) (stack_growup_p(x) ? a : b)
+#endif
+
+#define GC_WATER_MARK 512
 
 #define CHECK_STACK(ret) do {\
     SET_STACK_END;\
-    (ret) = (STACK_LENGTH > STACK_LEVEL_MAX);\
+    (ret) = (STACK_LENGTH > STACK_LEVEL_MAX + GC_WATER_MARK);\
 } while (0)
 
 int
@@ -391,7 +471,7 @@ ruby_stack_length(p)
     VALUE **p;
 {
     SET_STACK_END;
-    if (p) *p = STACK_END;
+    if (p) *p = STACK_UPPER(STACK_END, rb_gc_stack_start, STACK_END);
     return STACK_LENGTH;
 }
 
@@ -460,6 +540,9 @@ sweep_source_filename(key, value)
     }
 }
 
+static void gc_mark _((VALUE ptr, int lev));
+static void gc_mark_children _((VALUE ptr, int lev));
+
 static void
 gc_mark_all()
 {
@@ -468,11 +551,11 @@ gc_mark_all()
 
     init_mark_stack();
     for (i = 0; i < heaps_used; i++) {
-	p = heaps[i]; pend = p + heaps_limits[i];
+	p = heaps[i].slot; pend = p + heaps[i].limit;
 	while (p < pend) {
 	    if ((p->as.basic.flags & FL_MARK) &&
 		(p->as.basic.flags != FL_MARK)) {
-		rb_gc_mark((VALUE)p);
+		gc_mark_children((VALUE)p, 0);
 	    }
 	    p++;
 	}
@@ -491,7 +574,7 @@ gc_mark_rest()
     init_mark_stack();
     while(p != tmp_arry){
 	p--;
-	rb_gc_mark(*p);
+	gc_mark_children(*p, 0);
     }
 }
 
@@ -507,8 +590,8 @@ is_pointer_to_heap(ptr)
 
     /* check if p looks like a pointer */
     for (i=0; i < heaps_used; i++) {
-	heap_org = heaps[i];
-	if (heap_org <= p && p < heap_org + heaps_limits[i] &&
+	heap_org = heaps[i].slot;
+	if (heap_org <= p && p < heap_org + heaps[i].limit &&
 	    ((((char*)p)-((char*)heap_org))%sizeof(RVALUE)) == 0)
 	    return Qtrue;
     }
@@ -522,7 +605,7 @@ mark_locations_array(x, n)
 {
     while (n--) {
 	if (is_pointer_to_heap((void *)*x)) {
-	    rb_gc_mark(*x);
+	    gc_mark(*x, 0);
 	}
 	x++;
     }
@@ -532,51 +615,63 @@ void
 rb_gc_mark_locations(start, end)
     VALUE *start, *end;
 {
-    VALUE *tmp;
     long n;
 
-    if (start > end) {
-	tmp = start;
-	start = end;
-	end = tmp;
-    }
-    n = end - start + 1;
+    n = end - start;
     mark_locations_array(start,n);
 }
 
 static int
-mark_entry(key, value)
+mark_entry(key, value, lev)
     ID key;
     VALUE value;
+    int lev;
 {
-    rb_gc_mark(value);
+    gc_mark(value, lev);
     return ST_CONTINUE;
 }
 
+void
+mark_tbl(tbl, lev)
+    st_table *tbl;
+    int lev;
+{
+    if (!tbl) return;
+    st_foreach(tbl, mark_entry, lev+1);
+}
+
 void
 rb_mark_tbl(tbl)
     st_table *tbl;
 {
-    if (!tbl) return;
-    st_foreach(tbl, mark_entry, 0);
+    mark_tbl(tbl, 0);
 }
 
 static int
-mark_keyvalue(key, value)
+mark_keyvalue(key, value, lev)
     VALUE key;
     VALUE value;
+    int lev;
 {
-    rb_gc_mark(key);
-    rb_gc_mark(value);
+    gc_mark(key, lev);
+    gc_mark(value, lev);
     return ST_CONTINUE;
 }
 
+void
+mark_hash(tbl, lev)
+    st_table *tbl;
+    int lev;
+{
+    if (!tbl) return;
+    st_foreach(tbl, mark_keyvalue, lev+1);
+}
+
 void
 rb_mark_hash(tbl)
     st_table *tbl;
 {
-    if (!tbl) return;
-    st_foreach(tbl, mark_keyvalue, 0);
+    mark_hash(tbl, 0);
 }
 
 void
@@ -584,19 +679,26 @@ rb_gc_mark_maybe(obj)
     VALUE obj;
 {
     if (is_pointer_to_heap((void *)obj)) {
-	rb_gc_mark(obj);
+	gc_mark(obj, 0);
     }
 }
 
+#define GC_LEVEL_MAX 250
+
 void
-rb_gc_mark(ptr)
+gc_mark(ptr, lev)
     VALUE ptr;
+    int lev;
 {
-    int ret;
     register RVALUE *obj;
 
-    CHECK_STACK(ret);
-    if (ret) {
+    obj = RANY(ptr);
+    if (rb_special_const_p(ptr)) return; /* special const not marked */
+    if (obj->as.basic.flags == 0) return;       /* free cell */
+    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */ 
+    obj->as.basic.flags |= FL_MARK;
+
+    if (lev > GC_LEVEL_MAX) {
 	if (!mark_stack_overflow) {
 	    if (mark_stack_ptr - mark_stack < MARK_STACK_MAX) {
 		*mark_stack_ptr = ptr;
@@ -608,15 +710,33 @@ rb_gc_mark(ptr)
 	}
 	return;
     }
+    gc_mark_children(ptr, lev);
+}
+
+void
+rb_gc_mark(ptr)
+    VALUE ptr;
+{
+    gc_mark(ptr, 0);
+}
+
+static void
+gc_mark_children(ptr, lev)
+    VALUE ptr;
+    int lev;
+{
+    register RVALUE *obj = RANY(ptr);
+
+    goto marking;		/* skip */
 
   again:
     obj = RANY(ptr);
     if (rb_special_const_p(ptr)) return; /* special const not marked */
     if (obj->as.basic.flags == 0) return;       /* free cell */
     if (obj->as.basic.flags & FL_MARK) return;  /* already marked */ 
-
     obj->as.basic.flags |= FL_MARK;
 
+  marking:
     if (FL_TEST(obj, FL_EXIVAR)) {
 	rb_mark_generic_ivar(ptr);
     }
@@ -639,7 +759,7 @@ rb_gc_mark(ptr)
 	  case NODE_RESCUE:
 	  case NODE_RESBODY:
 	  case NODE_CLASS:
-	    rb_gc_mark((VALUE)obj->as.node.u2.node);
+	    gc_mark((VALUE)obj->as.node.u2.node, lev);
 	    /* fall through */
 	  case NODE_BLOCK:	/* 1,3 */
 	  case NODE_ARRAY:
@@ -652,12 +772,11 @@ rb_gc_mark(ptr)
 	  case NODE_CALL:
 	  case NODE_DEFS:
 	  case NODE_OP_ASGN1:
-	    rb_gc_mark((VALUE)obj->as.node.u1.node);
+	    gc_mark((VALUE)obj->as.node.u1.node, lev);
 	    /* fall through */
 	  case NODE_SUPER:	/* 3 */
 	  case NODE_FCALL:
 	  case NODE_DEFN:
-	  case NODE_NEWLINE:
 	    ptr = (VALUE)obj->as.node.u3.node;
 	    goto again;
 
@@ -676,7 +795,7 @@ rb_gc_mark(ptr)
 	  case NODE_OP_ASGN_OR:
 	  case NODE_OP_ASGN_AND:
 	  case NODE_MODULE:
-	    rb_gc_mark((VALUE)obj->as.node.u1.node);
+	    gc_mark((VALUE)obj->as.node.u1.node, lev);
 	    /* fall through */
 	  case NODE_METHOD:	/* 2 */
 	  case NODE_NOT:
@@ -705,13 +824,16 @@ rb_gc_mark(ptr)
 	  case NODE_YIELD:
 	  case NODE_COLON2:
 	  case NODE_ARGS:
+	  case NODE_SPLAT:
+	  case NODE_TO_ARY:
+	  case NODE_SVALUE:
 	    ptr = (VALUE)obj->as.node.u1.node;
 	    goto again;
 
 	  case NODE_SCOPE:	/* 2,3 */
 	  case NODE_BLOCK_PASS:
 	  case NODE_CDECL:
-	    rb_gc_mark((VALUE)obj->as.node.u3.node);
+	    gc_mark((VALUE)obj->as.node.u3.node, lev);
 	    ptr = (VALUE)obj->as.node.u2.node;
 	    goto again;
 
@@ -735,6 +857,7 @@ rb_gc_mark(ptr)
 	  case NODE_NIL:
 	  case NODE_TRUE:
 	  case NODE_FALSE:
+	  case NODE_ERRINFO:
 	  case NODE_ATTRSET:
 	  case NODE_BLOCK_ARG:
 	  case NODE_POSTEXE:
@@ -749,25 +872,25 @@ rb_gc_mark(ptr)
 
 	  default:		/* unlisted NODE */
 	    if (is_pointer_to_heap(obj->as.node.u1.node)) {
-		rb_gc_mark((VALUE)obj->as.node.u1.node);
+		gc_mark((VALUE)obj->as.node.u1.node, lev);
 	    }
 	    if (is_pointer_to_heap(obj->as.node.u2.node)) {
-		rb_gc_mark((VALUE)obj->as.node.u2.node);
+		gc_mark((VALUE)obj->as.node.u2.node, lev);
 	    }
 	    if (is_pointer_to_heap(obj->as.node.u3.node)) {
-		rb_gc_mark((VALUE)obj->as.node.u3.node);
+		gc_mark((VALUE)obj->as.node.u3.node, lev);
 	    }
 	}
 	return;			/* no need to mark class. */
     }
 
-    rb_gc_mark(obj->as.basic.klass);
+    gc_mark(obj->as.basic.klass, lev);
     switch (obj->as.basic.flags & T_MASK) {
       case T_ICLASS:
       case T_CLASS:
       case T_MODULE:
-	rb_mark_tbl(obj->as.klass.m_tbl);
-	rb_mark_tbl(obj->as.klass.iv_tbl);
+	mark_tbl(obj->as.klass.m_tbl, lev);
+	mark_tbl(obj->as.klass.iv_tbl, lev);
 	ptr = obj->as.klass.super;
 	goto again;
 
@@ -781,13 +904,13 @@ rb_gc_mark(ptr)
 	    VALUE *ptr = obj->as.array.ptr;
 
 	    for (i=0; i < len; i++) {
-		rb_gc_mark(*ptr++);
+		gc_mark(*ptr++, lev);
 	    }
 	}
 	break;
 
       case T_HASH:
-	rb_mark_hash(obj->as.hash.tbl);
+	mark_hash(obj->as.hash.tbl, lev);
 	ptr = obj->as.hash.ifnone;
 	goto again;
 
@@ -804,7 +927,7 @@ rb_gc_mark(ptr)
 	break;
 
       case T_OBJECT:
-	rb_mark_tbl(obj->as.object.iv_tbl);
+	mark_tbl(obj->as.object.iv_tbl, lev);
 	break;
 
       case T_FILE:
@@ -822,7 +945,7 @@ rb_gc_mark(ptr)
 	break;
 
       case T_VARMAP:
-	rb_gc_mark(obj->as.varmap.val);
+	gc_mark(obj->as.varmap.val, lev);
 	ptr = (VALUE)obj->as.varmap.next;
 	goto again;
 
@@ -832,7 +955,7 @@ rb_gc_mark(ptr)
 	    VALUE *vars = &obj->as.scope.local_vars[-1];
 
 	    while (n--) {
-		rb_gc_mark(*vars++);
+		gc_mark(*vars++, lev);
 	    }
 	}
 	break;
@@ -843,7 +966,7 @@ rb_gc_mark(ptr)
 	    VALUE *ptr = obj->as.rstruct.ptr;
 
 	    while (len--) {
-		rb_gc_mark(*ptr++);
+		gc_mark(*ptr++, lev);
 	    }
 	}
 	break;
@@ -857,15 +980,6 @@ rb_gc_mark(ptr)
 
 static void obj_free _((VALUE));
 
-static unsigned long
-size_of_table(tbl)
-    struct st_table *tbl;
-{
-    if (!tbl) return 0;
-    return tbl->num_bins * sizeof(struct st_table_entry *) +
-	tbl->num_entries * 4 * sizeof(VALUE);
-}
-
 static void
 gc_sweep()
 {
@@ -878,10 +992,10 @@ gc_sweep()
 	/* should not reclaim nodes during compilation
            if yacc's semantic stack is not allocated on machine stack */
 	for (i = 0; i < heaps_used; i++) {
-	    p = heaps[i]; pend = p + heaps_limits[i];
+	    p = heaps[i].slot; pend = p + heaps[i].limit;
 	    while (p < pend) {
 		if (!(p->as.basic.flags&FL_MARK) && BUILTIN_TYPE(p) == T_NODE)
-		    rb_gc_mark((VALUE)p);
+		    gc_mark((VALUE)p, 0);
 		p++;
 	    }
 	}
@@ -898,7 +1012,7 @@ gc_sweep()
 	RVALUE *free = freelist;
 	RVALUE *final = final_list;
 
-	p = heaps[i]; pend = p + heaps_limits[i];
+	p = heaps[i].slot; pend = p + heaps[i].limit;
 	while (p < pend) {
 	    if (!(p->as.basic.flags & FL_MARK)) {
 		if (p->as.basic.flags) {
@@ -926,10 +1040,10 @@ gc_sweep()
 	    }
 	    p++;
 	}
-	if (n == heaps_limits[i] && freed + n > FREE_MIN) {
+	if (n == heaps[i].limit && freed > FREE_MIN) {
 	    RVALUE *pp;
 
-	    heaps_limits[i] = 0;
+	    heaps[i].limit = 0;
 	    for (pp = final_list; pp != final; pp = pp->as.free.next) {
 		p->as.free.flags |= FL_SINGLETON; /* freeing page mark */
 	    }
@@ -939,8 +1053,10 @@ gc_sweep()
 	    freed += n;
 	}
     }
-    malloc_limit += (malloc_increase - malloc_limit) * (double)live / (live + freed);
-    if (malloc_limit < GC_MALLOC_LIMIT) malloc_limit = GC_MALLOC_LIMIT;
+    if (malloc_increase > malloc_limit) {
+	malloc_limit += (malloc_increase - malloc_limit) * (double)live / (live + freed);
+	if (malloc_limit < GC_MALLOC_LIMIT) malloc_limit = GC_MALLOC_LIMIT;
+    }
     malloc_increase = 0;
     if (freed < FREE_MIN) {
 	add_heap();
@@ -967,14 +1083,13 @@ gc_sweep()
 	}
     }
     for (i = j = 1; j < heaps_used; i++) {
-	if (heaps_limits[i] == 0) {
-	    free(heaps[i]);
+	if (heaps[i].limit == 0) {
+	    free(heaps[i].slot);
 	    heaps_used--;
 	}
 	else {
 	    if (i != j) {
 		heaps[j] = heaps[i];
-		heaps_limits[j] = heaps_limits[i];
 	    }
 	    j++;
 	}
@@ -1015,6 +1130,7 @@ obj_free(obj)
 	break;
       case T_MODULE:
       case T_CLASS:
+	rb_clear_cache_by_class((VALUE)obj);
 	st_free_table(RANY(obj)->as.klass.m_tbl);
 	if (RANY(obj)->as.object.iv_tbl) {
 	    st_free_table(RANY(obj)->as.object.iv_tbl);
@@ -1122,8 +1238,7 @@ rb_gc_mark_frame(frame)
     struct FRAME *frame;
 {
     mark_locations_array(frame->argv, frame->argc);
-    rb_gc_mark(frame->cbase);
-    rb_gc_mark((VALUE)frame->node);
+    gc_mark((VALUE)frame->node, 0);
 }
 
 #ifdef __GNUC__
@@ -1172,6 +1287,11 @@ rb_gc()
     jmp_buf save_regs_gc_mark;
     SET_STACK_END;
 
+#ifdef HAVE_NATIVETHREAD
+    if (!is_ruby_native_thread()) {
+	rb_bug("cross-thread violation on rb_gc()");
+    }
+#endif
     if (dont_gc || during_gc) {
 	if (!freelist) {
 	    add_heap();
@@ -1194,27 +1314,53 @@ rb_gc()
 	    }
 	}
     }
-    rb_gc_mark((VALUE)ruby_class);
-    rb_gc_mark((VALUE)ruby_scope);
-    rb_gc_mark((VALUE)ruby_dyna_vars);
+    gc_mark((VALUE)ruby_scope, 0);
+    gc_mark((VALUE)ruby_dyna_vars, 0);
     if (finalizer_table) {
-	rb_mark_tbl(finalizer_table);
+	mark_tbl(finalizer_table, 0);
     }
 
     FLUSH_REGISTER_WINDOWS;
-    /* This assumes that all registers are saved into the jmp_buf */
+    /* This assumes that all registers are saved into the jmp_buf (and stack) */
     setjmp(save_regs_gc_mark);
     mark_locations_array((VALUE*)save_regs_gc_mark, sizeof(save_regs_gc_mark) / sizeof(VALUE *));
-    rb_gc_mark_locations(rb_gc_stack_start, (VALUE*)STACK_END);
-#if defined(__human68k__)
-    rb_gc_mark_locations((VALUE*)((char*)rb_gc_stack_start + 2),
-			 (VALUE*)((char*)STACK_END + 2));
+#if STACK_GROW_DIRECTION < 0
+    rb_gc_mark_locations((VALUE*)STACK_END, rb_gc_stack_start);
+#elif STACK_GROW_DIRECTION > 0
+    rb_gc_mark_locations(rb_gc_stack_start, (VALUE*)STACK_END + 1);
+#else
+    if ((VALUE*)STACK_END < rb_gc_stack_start)
+	rb_gc_mark_locations((VALUE*)STACK_END, rb_gc_stack_start);
+    else
+	rb_gc_mark_locations(rb_gc_stack_start, (VALUE*)STACK_END + 1);
+#endif
+#ifdef __ia64__
+    /* mark backing store (flushed register window on the stack) */
+    /* the basic idea from guile GC code                         */
+    {
+	ucontext_t ctx;
+	VALUE *top, *bot;
+	getcontext(&ctx);
+	mark_locations_array((VALUE*)&ctx.uc_mcontext,
+			     ((size_t)(sizeof(VALUE)-1 + sizeof ctx.uc_mcontext)/sizeof(VALUE)));
+	bot = (VALUE*)__libc_ia64_register_backing_store_base;
+#if defined(__FreeBSD__)
+	top = (VALUE*)ctx.uc_mcontext.mc_special.bspstore;
+#else
+	top = (VALUE*)ctx.uc_mcontext.sc_ar_bsp;
+#endif
+	rb_gc_mark_locations(bot, top);
+    }
+#endif
+#if defined(__human68k__) || defined(__mc68000__)
+    rb_gc_mark_locations((VALUE*)((char*)STACK_END + 2),
+			 (VALUE*)((char*)rb_gc_stack_start + 2));
 #endif
     rb_gc_mark_threads();
 
     /* mark protected global variables */
     for (list = global_List; list; list = list->next) {
-	rb_gc_mark(*list->varptr);
+	rb_gc_mark_maybe(*list->varptr);
     }
     rb_mark_end_proc();
     rb_gc_mark_global_tbl();
@@ -1239,6 +1385,16 @@ rb_gc()
     gc_sweep();
 }
 
+/*
+ *  call-seq:
+ *     GC.start                     => nil
+ *     gc.garbage_collect           => nil
+ *     ObjectSpace.garbage_collect  => nil
+ *
+ *  Initiates garbage collection, unless manually disabled.
+ *     
+ */
+
 VALUE
 rb_gc_start()
 {
@@ -1246,18 +1402,6 @@ rb_gc_start()
     return Qnil;
 }
 
-#if !defined(__human68k__)
-static int
-stack_growup_p(addr)
-    VALUE *addr;
-{
-    SET_STACK_END;
-
-    if (STACK_END > addr) return Qtrue;
-    return Qfalse;
-}
-#endif
-
 void
 Init_stack(addr)
     VALUE *addr;
@@ -1268,14 +1412,10 @@ Init_stack(addr)
 #else
     if (!addr) addr = (VALUE *)&addr;
     if (rb_gc_stack_start) {
-	if (stack_growup_p(addr)) {
-	    if (rb_gc_stack_start > addr)
-		rb_gc_stack_start = addr;
-	}
-	else {
-	    if (rb_gc_stack_start < addr)
-		rb_gc_stack_start = addr;
-	}
+	if (STACK_UPPER(&addr,
+			rb_gc_stack_start > addr,
+			rb_gc_stack_start < ++addr))
+	    rb_gc_stack_start = addr;
 	return;
     }
     rb_gc_stack_start = addr;
@@ -1285,15 +1425,55 @@ Init_stack(addr)
 	struct rlimit rlim;
 
 	if (getrlimit(RLIMIT_STACK, &rlim) == 0) {
-	    double space = (double)rlim.rlim_cur*0.2;
+	    unsigned int space = rlim.rlim_cur/5;
 
 	    if (space > 1024*1024) space = 1024*1024;
 	    STACK_LEVEL_MAX = (rlim.rlim_cur - space) / sizeof(VALUE);
 	}
     }
+#if defined(__ia64__) && (!defined(__GNUC__) || __GNUC__ < 2 || defined(__OPTIMIZE__))
+    /* ruby crashes on IA64 if compiled with optimizer on */
+    /* when if STACK_LEVEL_MAX is greater than this magic number */
+    /* I know this is a kludge.  I suspect optimizer bug */
+#define IA64_MAGIC_STACK_LIMIT 49152
+    if (STACK_LEVEL_MAX > IA64_MAGIC_STACK_LIMIT)
+	STACK_LEVEL_MAX = IA64_MAGIC_STACK_LIMIT;
+#endif
 #endif
 }
 
+
+/*
+ * Document-class: ObjectSpace
+ *
+ *  The <code>ObjectSpace</code> module contains a number of routines
+ *  that interact with the garbage collection facility and allow you to
+ *  traverse all living objects with an iterator.
+ *     
+ *  <code>ObjectSpace</code> also provides support for object
+ *  finalizers, procs that will be called when a specific object is
+ *  about to be destroyed by garbage collection.
+ *     
+ *     include ObjectSpace
+ *     
+ *     
+ *     a = "A"
+ *     b = "B"
+ *     c = "C"
+ *     
+ *     
+ *     define_finalizer(a, proc {|id| puts "Finalizer one on #{id}" })
+ *     define_finalizer(a, proc {|id| puts "Finalizer two on #{id}" })
+ *     define_finalizer(b, proc {|id| puts "Finalizer three on #{id}" })
+ *     
+ *  <em>produces:</em>
+ *     
+ *     Finalizer three on 537763470
+ *     Finalizer one on 537763480
+ *     Finalizer two on 537763480
+ *     
+ */
+
 void
 Init_heap()
 {
@@ -1312,7 +1492,7 @@ os_live_obj()
     for (i = 0; i < heaps_used; i++) {
 	RVALUE *p, *pend;
 
-	p = heaps[i]; pend = p + heaps_limits[i];
+	p = heaps[i].slot; pend = p + heaps[i].limit;
 	for (;p < pend; p++) {
 	    if (p->as.basic.flags) {
 		switch (TYPE(p)) {
@@ -1345,7 +1525,7 @@ os_obj_of(of)
     for (i = 0; i < heaps_used; i++) {
 	RVALUE *p, *pend;
 
-	p = heaps[i]; pend = p + heaps_limits[i];
+	p = heaps[i].slot; pend = p + heaps[i].limit;
 	for (;p < pend; p++) {
 	    if (p->as.basic.flags) {
 		switch (TYPE(p)) {
@@ -1370,6 +1550,39 @@ os_obj_of(of)
     return INT2FIX(n);
 }
 
+/*
+ *  call-seq:
+ *     ObjectSpace.each_object([module]) {|obj| ... } => fixnum
+ *  
+ *  Calls the block once for each living, nonimmediate object in this
+ *  Ruby process. If <i>module</i> is specified, calls the block
+ *  for only those classes or modules that match (or are a subclass of)
+ *  <i>module</i>. Returns the number of objects found. Immediate
+ *  objects (<code>Fixnum</code>s, <code>Symbol</code>s
+ *  <code>true</code>, <code>false</code>, and <code>nil</code>) are
+ *  never returned. In the example below, <code>each_object</code>
+ *  returns both the numbers we defined and several constants defined in
+ *  the <code>Math</code> module.
+ *     
+ *     a = 102.7
+ *     b = 95       # Won't be returned
+ *     c = 12345678987654321
+ *     count = ObjectSpace.each_object(Numeric) {|x| p x }
+ *     puts "Total count: #{count}"
+ *     
+ *  <em>produces:</em>
+ *     
+ *     12345678987654321
+ *     102.7
+ *     2.71828182845905
+ *     3.14159265358979
+ *     2.22044604925031e-16
+ *     1.7976931348623157e+308
+ *     2.2250738585072e-308
+ *     Total count: 7
+ *     
+ */
+
 static VALUE
 os_each_obj(argc, argv)
     int argc;
@@ -1377,6 +1590,7 @@ os_each_obj(argc, argv)
 {
     VALUE of;
 
+    rb_secure(4);
     if (rb_scan_args(argc, argv, "01", &of) == 0) {
 	return os_live_obj();
     }
@@ -1387,28 +1601,37 @@ os_each_obj(argc, argv)
 
 static VALUE finalizers;
 
+/* deprecated
+ */
+
 static VALUE
-add_final(os, proc)
-    VALUE os, proc;
+add_final(os, block)
+    VALUE os, block;
 {
     rb_warn("ObjectSpace::add_finalizer is deprecated; use define_finalizer");
-    if (!rb_obj_is_kind_of(proc, rb_cProc)) {
-	rb_raise(rb_eArgError, "wrong type argument %s (Proc required)",
-		 rb_obj_classname(proc));
+    if (!rb_respond_to(block, rb_intern("call"))) {
+	rb_raise(rb_eArgError, "wrong type argument %s (should be callable)",
+		 rb_obj_classname(block));
     }
-    rb_ary_push(finalizers, proc);
-    return proc;
+    rb_ary_push(finalizers, block);
+    return block;
 }
 
+/*
+ * deprecated
+ */
 static VALUE
-rm_final(os, proc)
-    VALUE os, proc;
+rm_final(os, block)
+    VALUE os, block;
 {
     rb_warn("ObjectSpace::remove_finalizer is deprecated; use undefine_finalizer");
-    rb_ary_delete(finalizers, proc);
-    return proc;
+    rb_ary_delete(finalizers, block);
+    return block;
 }
 
+/*
+ * deprecated
+ */
 static VALUE
 finals()
 {
@@ -1416,6 +1639,10 @@ finals()
     return finalizers;
 }
 
+/*
+ * deprecated
+ */
+
 static VALUE
 call_final(os, obj)
     VALUE os, obj;
@@ -1426,31 +1653,48 @@ call_final(os, obj)
     return obj;
 }
 
+/*
+ *  call-seq:
+ *     ObjectSpace.undefine_finalizer(obj)
+ *  
+ *  Removes all finalizers for <i>obj</i>.
+ *     
+ */
+
 static VALUE
 undefine_final(os, obj)
     VALUE os, obj;
 {
     if (finalizer_table) {
-	st_delete(finalizer_table, &obj, 0);
+	st_delete(finalizer_table, (st_data_t*)&obj, 0);
     }
     return obj;
 }
 
+/*
+ *  call-seq:
+ *     ObjectSpace.define_finalizer(obj, aProc=proc())
+ *  
+ *  Adds <i>aProc</i> as a finalizer, to be called when <i>obj</i>
+ *  is about to be destroyed.
+ *     
+ */
+
 static VALUE
 define_final(argc, argv, os)
     int argc;
     VALUE *argv;
     VALUE os;
 {
-    VALUE obj, proc, table;
+    VALUE obj, block, table;
 
-    rb_scan_args(argc, argv, "11", &obj, &proc);
+    rb_scan_args(argc, argv, "11", &obj, &block);
     if (argc == 1) {
-	proc = rb_f_lambda();
+	block = rb_block_proc();
     }
-    else if (!rb_obj_is_kind_of(proc, rb_cProc)) {
-	rb_raise(rb_eArgError, "wrong type argument %s (Proc required)",
-		 rb_obj_classname(proc));
+    else if (!rb_respond_to(block, rb_intern("call"))) {
+	rb_raise(rb_eArgError, "wrong type argument %s (should be callable)",
+		 rb_obj_classname(block));
     }
     need_call_final = 1;
     FL_SET(obj, FL_FINALIZE);
@@ -1459,12 +1703,12 @@ define_final(argc, argv, os)
 	finalizer_table = st_init_numtable();
     }
     if (st_lookup(finalizer_table, obj, &table)) {
-	rb_ary_push(table, proc);
+	rb_ary_push(table, block);
     }
     else {
-	st_add_direct(finalizer_table, obj, rb_ary_new3(1, proc));
+	st_add_direct(finalizer_table, obj, rb_ary_new3(1, block));
     }
-    return proc;
+    return block;
 }
 
 void
@@ -1496,20 +1740,22 @@ run_final(obj)
     VALUE obj;
 {
     long i;
-    int status;
+    int status, critical_save = rb_thread_critical;
     VALUE args[2], table;
 
+    rb_thread_critical = Qtrue;
     args[1] = rb_ary_new3(1, rb_obj_id(obj)); /* make obj into id */
     for (i=0; i<RARRAY(finalizers)->len; i++) {
 	args[0] = RARRAY(finalizers)->ptr[i];
 	rb_protect((VALUE(*)_((VALUE)))run_single_final, (VALUE)args, &status);
     }
-    if (finalizer_table && st_delete(finalizer_table, &obj, &table)) {
+    if (finalizer_table && st_delete(finalizer_table, (st_data_t*)&obj, &table)) {
 	for (i=0; i<RARRAY(table)->len; i++) {
 	    args[0] = RARRAY(table)->ptr[i];
 	    rb_protect((VALUE(*)_((VALUE)))run_single_final, (VALUE)args, &status);
 	}
     }
+    rb_thread_critical = critical_save;
 }
 
 void
@@ -1529,7 +1775,7 @@ rb_gc_call_finalizer_at_exit()
 	    }
 	}
 	for (i = 0; i < heaps_used; i++) {
-	    p = heaps[i]; pend = p + heaps_limits[i];
+	    p = heaps[i].slot; pend = p + heaps[i].limit;
 	    while (p < pend) {
 		if (FL_TEST(p, FL_FINALIZE)) {
 		    FL_UNSET(p, FL_FINALIZE);
@@ -1540,14 +1786,19 @@ rb_gc_call_finalizer_at_exit()
 	    }
 	}
     }
-    /* run data object's finaliers */
+    /* run data object's finalizers */
     for (i = 0; i < heaps_used; i++) {
-	p = heaps[i]; pend = p + heaps_limits[i];
+	p = heaps[i].slot; pend = p + heaps[i].limit;
 	while (p < pend) {
 	    if (BUILTIN_TYPE(p) == T_DATA &&
 		DATA_PTR(p) && RANY(p)->as.data.dfree) {
 		p->as.free.flags = 0;
-		(*RANY(p)->as.data.dfree)(DATA_PTR(p));
+		if ((long)RANY(p)->as.data.dfree == -1) {
+		    RUBY_CRITICAL(free(DATA_PTR(p)));
+		}
+		else if (RANY(p)->as.data.dfree) {
+		    (*RANY(p)->as.data.dfree)(DATA_PTR(p));
+		}
 	    }
 	    else if (BUILTIN_TYPE(p) == T_FILE) {
 		p->as.free.flags = 0;
@@ -1558,6 +1809,19 @@ rb_gc_call_finalizer_at_exit()
     }
 }
 
+/*
+ *  call-seq:
+ *     ObjectSpace._id2ref(object_id) -> an_object
+ *  
+ *  Converts an object id to a reference to the object. May not be
+ *  called on an object id passed as a parameter to a finalizer.
+ *     
+ *     s = "I am a string"                    #=> "I am a string"
+ *     r = ObjectSpace._id2ref(s.object_id)   #=> "I am a string"
+ *     r == s                                 #=> true
+ *     
+ */
+
 static VALUE
 id2ref(obj, id)
     VALUE obj, id;
@@ -1575,15 +1839,21 @@ id2ref(obj, id)
     }
 
     ptr = id ^ FIXNUM_FLAG;	/* unset FIXNUM_FLAG */
-    if (!is_pointer_to_heap((void *)ptr)) {
+    if (!is_pointer_to_heap((void *)ptr)|| BUILTIN_TYPE(ptr) >= T_BLKTAG) {
 	rb_raise(rb_eRangeError, "0x%lx is not id value", p0);
     }
-    if (BUILTIN_TYPE(ptr) == 0) {
+    if (BUILTIN_TYPE(ptr) == 0 || RBASIC(ptr)->klass == 0) {
 	rb_raise(rb_eRangeError, "0x%lx is recycled object", p0);
     }
     return (VALUE)ptr;
 }
 
+/*
+ *  The <code>GC</code> module provides an interface to Ruby's mark and
+ *  sweep garbage collection mechanism. Some of the underlying methods
+ *  are also available via the <code>ObjectSpace</code> module.
+ */
+
 void
 Init_GC()
 {
