@@ -2,8 +2,8 @@
 
   ruby.c -
 
-  $Author: nobu $
-  $Date: 2003/02/13 03:00:55 $
+  $Author: matz $
+  $Date: 2004/01/21 16:47:14 $
   created at: Tue Aug 10 12:47:31 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -17,6 +17,7 @@
 #endif
 #ifdef _WIN32_WCE
 #include <winsock.h>
+#include "wince.h"
 #endif
 #include "ruby.h"
 #include "dln.h"
@@ -93,6 +94,7 @@ usage(name)
 "-T[level]       turn on tainting checks",
 "-v              print version number, then turn on verbose mode",
 "-w              turn warnings on for your script",
+"-W[level]       set warning level; 0=silence, 1=medium, 2=verbose (default)",
 "-x[directory]   strip off text before #!ruby line and perhaps cd to directory",
 "--copyright     print the copyright",
 "--version       print the version",
@@ -210,7 +212,7 @@ ruby_incpush(path)
 #define LOAD_RELATIVE 1
 #endif
 
-#ifdef DOSISH
+#if defined DOSISH || defined __CYGWIN__
 static inline void translate_char _((char *, int, int));
 
 static inline void
@@ -224,7 +226,7 @@ translate_char(p, from, to)
 #ifdef CharNext		/* defined as CharNext[AW] on Windows. */
 	p = CharNext(p);
 #else
-	p += mblen(p, MB_CUR_MAX);
+	p += mblen(p, RUBY_MBCHAR_MAXSIZE);
 #endif
     }
 }
@@ -241,9 +243,11 @@ ruby_init_loadpath()
     HMODULE libruby = NULL;
     MEMORY_BASIC_INFORMATION m;
 
+#ifndef _WIN32_WCE
     memset(&m, 0, sizeof(m));
     if (VirtualQuery(ruby_init_loadpath, &m, sizeof(m)) && m.State == MEM_COMMIT)
 	libruby = (HMODULE)m.AllocationBase;
+#endif
     GetModuleFileName(libruby, libpath, sizeof libpath);
 #elif defined(DJGPP)
     extern char *__dos_argv0;
@@ -256,7 +260,7 @@ ruby_init_loadpath()
 #endif
 
     libpath[FILENAME_MAX] = '\0';
-#ifdef DOSISH
+#if defined DOSISH || defined __CYGWIN__
     translate_char(libpath, '\\', '/');
 #endif
     p = strrchr(libpath, '/');
@@ -339,7 +343,7 @@ require_libraries()
     Init_ext();		/* should be called here for some reason :-( */
     save[0] = ruby_eval_tree;
     save[1] = ruby_eval_tree_begin;
-    save[2] = NEW_NEWLINE(0);
+    save[2] = NEW_BEGIN(0);
     ruby_eval_tree = ruby_eval_tree_begin = 0;
     req_list_last = 0;
     while (list) {
@@ -425,6 +429,7 @@ proc_options(argc, argv)
     char *argv0 = argv[0];
     int do_search;
     char *s;
+    NODE *volatile script_node = 0;
 
     int version = 0;
     int copyright = 0;
@@ -477,6 +482,27 @@ proc_options(argc, argv)
 	    s++;
 	    goto reswitch;
 
+	  case 'W':
+	    {
+		int numlen;
+		int v = 2;	/* -W as -W2 */
+
+		if (*++s) {
+		    v = scan_oct(s, 1, &numlen);
+		    if (numlen == 0) v = 1;
+		    s += numlen;
+		}
+		switch (v) {
+		  case 0:
+		    ruby_verbose = Qnil; break;
+		  case 1:
+		    ruby_verbose = Qfalse; break;
+		  default:
+		    ruby_verbose = Qtrue; break;
+		}
+	    }
+	    goto reswitch;
+
 	  case 'c':
 	    do_check = Qtrue;
 	    s++;
@@ -564,7 +590,7 @@ proc_options(argc, argv)
 
 	  case 'F':
 	    if (*++s) {
-		rb_fs = rb_str_new2(s);
+		rb_fs = rb_reg_new(s, strlen(s), 0);
 	    }
 	    break;
 
@@ -701,8 +727,8 @@ proc_options(argc, argv)
     }
 
     if (rb_safe_level() >= 4) {
-      OBJ_TAINT(rb_argv);
-      OBJ_TAINT(rb_load_path);
+	OBJ_TAINT(rb_argv);
+	OBJ_TAINT(rb_load_path);
     }
 
     if (!e_script) {
@@ -723,11 +749,13 @@ proc_options(argc, argv)
 		    script = dln_find_file(argv[0], path);
 		}
 		if (!script) {
-		    script = dln_find_file(argv[0], getenv("PATH"));
+		    script = dln_find_file(argv[0], getenv(PATH_ENV));
 		}
 		if (!script) script = argv[0];
+		script = ruby_sourcefile = rb_source_filename(script);
+		script_node = NEW_BEGIN(0);
 	    }
-#ifdef DOSISH
+#if defined DOSISH || defined __CYGWIN__
 	    translate_char(script, '\\', '/');
 #endif
 	    argc--; argv++;
@@ -755,8 +783,8 @@ proc_options(argc, argv)
     xflag = 0;
 
     if (rb_safe_level() >= 4) {
-      FL_UNSET(rb_argv, FL_TAINT);
-      FL_UNSET(rb_load_path, FL_TAINT);
+	FL_UNSET(rb_argv, FL_TAINT);
+	FL_UNSET(rb_load_path, FL_TAINT);
     }
 }
 
@@ -771,6 +799,7 @@ load_file(fname, script)
     VALUE f;
     int line_start = 1;
 
+    if (!fname) rb_load_fail(fname);
     if (strcmp(fname, "-") == 0) {
 	f = rb_stdin;
     }
@@ -1010,15 +1039,24 @@ forbid_setid(s)
         rb_raise(rb_eSecurityError, "No %s allowed in tainted mode", s);
 }
 
+static void
+verbose_setter(val, id, variable)
+    VALUE val;
+    ID id;
+    VALUE *variable;
+{
+    ruby_verbose = RTEST(val) ? Qtrue : val;
+}
+
 void
 ruby_prog_init()
 {
     init_ids();
 
     ruby_sourcefile = rb_source_filename("ruby");
-    rb_define_variable("$VERBOSE", &ruby_verbose);
-    rb_define_variable("$-v", &ruby_verbose);
-    rb_define_variable("$-w", &ruby_verbose);
+    rb_define_hooked_variable("$VERBOSE", &ruby_verbose, 0, verbose_setter);
+    rb_define_hooked_variable("$-v", &ruby_verbose, 0, verbose_setter);
+    rb_define_hooked_variable("$-w", &ruby_verbose, 0, verbose_setter);
     rb_define_variable("$DEBUG", &ruby_debug);
     rb_define_variable("$-d", &ruby_debug);
     rb_define_readonly_variable("$-p", &do_print);
@@ -1055,7 +1093,10 @@ ruby_set_argv(argc, argv)
 #endif
     rb_ary_clear(rb_argv);
     for (i=0; i < argc; i++) {
-	rb_ary_push(rb_argv, rb_tainted_str_new2(argv[i]));
+	VALUE arg = rb_tainted_str_new2(argv[i]);
+
+	OBJ_FREEZE(arg);
+	rb_ary_push(rb_argv, arg);
     }
 }
 
