@@ -3,7 +3,7 @@
   variable.c -
 
   $Author: matz $
-  $Date: 2003/02/18 14:30:17 $
+  $Date: 2004/01/19 09:19:29 $
   created at: Tue Apr 19 23:55:15 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -20,12 +20,16 @@
 
 static st_table *rb_global_tbl;
 st_table *rb_class_tbl;
+static ID autoload, classpath, tmp_classpath;
 
 void
 Init_var_tables()
 {
     rb_global_tbl = st_init_numtable();
     rb_class_tbl = st_init_numtable();
+    autoload = rb_intern("__autoload__");
+    classpath = rb_intern("__classpath__");
+    tmp_classpath = rb_intern("__tmp_classpath__");
 }
 
 struct fc_result {
@@ -47,7 +51,7 @@ fc_path(fc, name)
     while (fc) {
 	if (fc->track == rb_cObject) break;
 	if (ROBJECT(fc->track)->iv_tbl &&
-	    st_lookup(ROBJECT(fc->track)->iv_tbl, rb_intern("__classpath__"), &tmp)) {
+	    st_lookup(ROBJECT(fc->track)->iv_tbl, classpath, &tmp)) {
 	    tmp = rb_str_dup(tmp);
 	    rb_str_cat2(tmp, "::");
 	    rb_str_append(tmp, path);
@@ -104,7 +108,6 @@ fc_i(key, value, res)
 
       default:
 	break;
-	    
     }
     return ST_CONTINUE;
 }
@@ -129,8 +132,9 @@ find_class_path(klass)
     if (arg.path) {
 	if (!ROBJECT(klass)->iv_tbl) {
 	    ROBJECT(klass)->iv_tbl = st_init_numtable();
-	} 
-	st_insert(ROBJECT(klass)->iv_tbl,rb_intern("__classpath__"),arg.path);
+	}
+	st_insert(ROBJECT(klass)->iv_tbl, classpath, arg.path);
+	st_delete(RCLASS(klass)->iv_tbl, &tmp_classpath, 0);
 	return arg.path;
     }
     return Qnil;
@@ -141,39 +145,41 @@ classname(klass)
     VALUE klass;
 {
     VALUE path = Qnil;
-    ID classpath = rb_intern("__classpath__");
 
-    klass = rb_class_real(klass);
     if (!klass) klass = rb_cObject;
-    if (ROBJECT(klass)->iv_tbl &&
-	!st_lookup(ROBJECT(klass)->iv_tbl, classpath, &path)) {
-	ID classid = rb_intern("__classid__");
+    if (ROBJECT(klass)->iv_tbl) {
+	if (!st_lookup(ROBJECT(klass)->iv_tbl, classpath, &path)) {
+	    ID classid = rb_intern("__classid__");
 
-	if (st_lookup(ROBJECT(klass)->iv_tbl, classid, &path)) {
+	    if (!st_lookup(ROBJECT(klass)->iv_tbl, classid, &path)) {
+		return find_class_path(klass);
+	    }
 	    path = rb_str_new2(rb_id2name(SYM2ID(path)));
 	    st_insert(ROBJECT(klass)->iv_tbl, classpath, path);
-	    st_delete(RCLASS(klass)->iv_tbl, &classid, 0);
+	    st_delete(RCLASS(klass)->iv_tbl, (st_data_t*)&classid, 0);
 	}
-    }
-    if (NIL_P(path)) {
-	path = find_class_path(klass);
-	if (NIL_P(path)) {
-	    return 0;
+	if (TYPE(path) != T_STRING) {
+	    rb_bug("class path is not set properly");
 	}
 	return path;
     }
-    if (TYPE(path) != T_STRING)
-	rb_bug("class path is not set properly");
-    return path;
+    return find_class_path(klass);
 }
 
+/*
+ *  call-seq:
+ *     mod.name    => string
+ *  
+ *  Returns the name of the module <i>mod</i>.
+ */
+
 VALUE
 rb_mod_name(mod)
     VALUE mod;
 {
     VALUE path = classname(mod);
 
-    if (path) return rb_str_dup(path);
+    if (!NIL_P(path)) return rb_str_dup(path);
     return rb_str_new(0,0);
 }
 
@@ -183,9 +189,12 @@ rb_class_path(klass)
 {
     VALUE path = classname(klass);
 
-    if (path) return path;
+    if (!NIL_P(path)) return path;
+    if (RCLASS(klass)->iv_tbl && st_lookup(RCLASS(klass)->iv_tbl,
+					   tmp_classpath, &path)) {
+	return path;
+    }
     else {
-	VALUE str;
 	char *s = "Class";
 
 	if (TYPE(klass) == T_MODULE) {
@@ -196,11 +205,12 @@ rb_class_path(klass)
 		s = rb_class2name(RBASIC(klass)->klass);
 	    }
 	}
-	str = rb_str_new(0, 2 + strlen(s) + 3 + 2 * SIZEOF_LONG + 1);
-	sprintf(RSTRING(str)->ptr, "#<%s:0x%lx>", s, klass);
-	RSTRING(str)->len = strlen(RSTRING(str)->ptr);
+	path = rb_str_new(0, 2 + strlen(s) + 3 + 2 * SIZEOF_LONG + 1);
+	sprintf(RSTRING(path)->ptr, "#<%s:0x%lx>", s, klass);
+	RSTRING(path)->len = strlen(RSTRING(path)->ptr);
+	rb_ivar_set(klass, tmp_classpath, path);
 
-	return str;
+	return path;
     }
 }
 
@@ -219,7 +229,7 @@ rb_set_class_path(klass, under, name)
 	rb_str_cat2(str, "::");
 	rb_str_cat2(str, name);
     }
-    rb_iv_set(klass, "__classpath__", str);
+    rb_ivar_set(klass, classpath, str);
 }
 
 VALUE
@@ -247,8 +257,7 @@ rb_path2class(path)
 	}
 	if (!rb_const_defined(c, id)) {
 	  undefined_class:
-	    rb_raise(rb_eArgError, "undefined class/module %s", rb_id2name(id));
-	    rb_raise(rb_eArgError, "undefined class/module %s", path);
+	    rb_raise(rb_eArgError, "undefined class/module %.*s", p-path, path);
 	}
 	c = rb_const_get_at(c, id);
 	switch (TYPE(c)) {
@@ -271,44 +280,18 @@ rb_name_class(klass, id)
     rb_iv_set(klass, "__classid__", ID2SYM(id));
 }
 
-static st_table *autoload_tbl = 0;
-
-static void
-rb_autoload_id(id, filename)
-    ID id;
-    const char *filename;
-{
-    rb_secure(4);
-    if (!rb_is_const_id(id)) {
-	rb_name_error(id, "autoload must be constant name");
-    }
-
-    if (!autoload_tbl) {
-	autoload_tbl = st_init_numtable();
-    }
-    st_insert(autoload_tbl, id, (st_data_t)strdup(filename));
-}
-
-void
-rb_autoload(klass, filename)
-    const char *klass, *filename;
-{
-    rb_autoload_id(rb_intern(klass), filename);
-}
-
 VALUE
-rb_f_autoload(obj, klass, file)
-    VALUE obj, klass, file;
+rb_class_name(klass)
+    VALUE klass;
 {
-    rb_autoload_id(rb_to_id(klass), StringValuePtr(file));
-    return Qnil;
+    return rb_class_path(rb_class_real(klass));
 }
 
 char *
 rb_class2name(klass)
     VALUE klass;
 {
-    return RSTRING(rb_class_path(klass))->ptr;
+    return RSTRING(rb_class_name(klass))->ptr;
 }
 
 char *
@@ -461,7 +444,7 @@ readonly_setter(val, id, var)
     ID id;
     void *var;
 {
-    rb_name_error(id, "can't set variable %s", rb_id2name(id));
+    rb_name_error(id, "%s is a read-only variable", rb_id2name(id));
 }
 
 static int
@@ -554,6 +537,29 @@ rb_trace_eval(cmd, val)
     rb_eval_cmd(cmd, rb_ary_new3(1, val), 0);
 }
 
+/*
+ *  call-seq:
+ *     trace_var(symbol, cmd )             => nil
+ *     trace_var(symbol) {|val| block }    => nil
+ *  
+ *  Controls tracing of assignments to global variables. The parameter
+ *  +symbol_ identifies the variable (as either a string name or a
+ *  symbol identifier). _cmd_ (which may be a string or a
+ *  +Proc+ object) or block is executed whenever the variable
+ *  is assigned. The block or +Proc+ object receives the
+ *  variable's new value as a parameter. Also see
+ *  <code>Kernel::untrace_var</code>.
+ *     
+ *     trace_var :$_, proc {|v| puts "$_ is now '#{v}'" }
+ *     $_ = "hello"
+ *     $_ = ' there'
+ *     
+ *  <em>produces:</em>
+ *     
+ *     $_ is now 'hello'
+ *     $_ is now ' there'
+ */
+
 VALUE
 rb_f_trace_var(argc, argv)
     int argc;
@@ -565,7 +571,7 @@ rb_f_trace_var(argc, argv)
 
     rb_secure(4);
     if (rb_scan_args(argc, argv, "11", &var, &cmd) == 1) {
-	cmd = rb_f_lambda();
+	cmd = rb_block_proc();
     }
     if (NIL_P(cmd)) {
 	return rb_f_untrace_var(argc, argv);
@@ -607,6 +613,16 @@ remove_trace(var)
     var->trace = t.next;
 }
 
+/*
+ *  call-seq:
+ *     untrace_var(symbol [, cmd] )   => array or nil
+ *  
+ *  Removes tracing for the specified command on the given global
+ *  variable and returns +nil+. If no command is specified,
+ *  removes all tracing for that variable and returns an array
+ *  containing the commands actually removed.
+ */
+
 VALUE
 rb_f_untrace_var(argc, argv)
     int argc;
@@ -662,7 +678,7 @@ struct trace_data {
     struct trace_var *trace;
     VALUE val;
 };
-    
+
 static VALUE
 trace_ev(data)
     struct trace_data *data;
@@ -745,6 +761,15 @@ gvar_i(key, entry, ary)
     return ST_CONTINUE;
 }
 
+/*
+ *  call-seq:
+ *     global_variables    => array
+ *  
+ *  Returns an array of the names of global variables.
+ *     
+ *     global_variables.grep /std/   #=> ["$stderr", "$stdout", "$stdin"]
+ */
+
 VALUE
 rb_f_global_variables()
 {
@@ -810,6 +835,7 @@ rb_generic_ivar_table(obj)
 {
     st_table *tbl;
 
+    if (!FL_TEST(obj, FL_EXIVAR)) return 0;
     if (!generic_iv_tbl) return 0;
     if (!st_lookup(generic_iv_tbl, obj, (st_data_t *)&tbl)) return 0;
     return tbl;
@@ -1063,6 +1089,23 @@ ivar_i(key, entry, ary)
     return ST_CONTINUE;
 }
 
+/*
+ *  call-seq:
+ *     obj.instance_variables    => array
+ *  
+ *  Returns an array of instance variable names for the receiver. Note
+ *  that simply defining an accessor does not create the corresponding
+ *  instance variable.
+ *     
+ *     class Fred
+ *       attr_accessor :a1
+ *       def initialize
+ *         @iv = 3
+ *       end
+ *     end
+ *     Fred.new.instance_variables   #=> ["@iv"]
+ */
+
 VALUE
 rb_obj_instance_variables(obj)
     VALUE obj;
@@ -1092,6 +1135,28 @@ rb_obj_instance_variables(obj)
     return ary;
 }
 
+/*
+ *  call-seq:
+ *     obj.remove_instance_variable(symbol)    => obj
+ *  
+ *  Removes the named instance variable from <i>obj</i>, returning that
+ *  variable's value.
+ *     
+ *     class Dummy
+ *       attr_reader :var
+ *       def initialize
+ *         @var = 99
+ *       end
+ *       def remove
+ *         remove_instance_variable(:@var)
+ *       end
+ *     end
+ *     d = Dummy.new
+ *     d.var      #=> 99
+ *     d.remove   #=> 99
+ *     d.var      #=> nil
+ */
+
 VALUE
 rb_obj_remove_instance_variable(obj, name)
     VALUE obj, name;
@@ -1103,14 +1168,14 @@ rb_obj_remove_instance_variable(obj, name)
 	rb_raise(rb_eSecurityError, "Insecure: can't modify instance variable");
     if (OBJ_FROZEN(obj)) rb_error_frozen("object");
     if (!rb_is_instance_id(id)) {
-	rb_name_error(id, "`%s' is not an instance variable", rb_id2name(id));
+	rb_name_error(id, "`%s' is not allowed as an instance variable name", rb_id2name(id));
     }
 
     switch (TYPE(obj)) {
       case T_OBJECT:
       case T_CLASS:
       case T_MODULE:
-	if (ROBJECT(obj)->iv_tbl && st_delete(ROBJECT(obj)->iv_tbl, &id, &val)) {
+	if (ROBJECT(obj)->iv_tbl && st_delete(ROBJECT(obj)->iv_tbl, (st_data_t*)&id, &val)) {
 	    return val;
 	}
 	break;
@@ -1126,64 +1191,209 @@ rb_obj_remove_instance_variable(obj, name)
     return Qnil;		/* not reached */
 }
 
-static int
-top_const_get(id, klassp)
-    ID id;
-    VALUE *klassp;
-{
-    /* pre-defined class */
-    if (st_lookup(rb_class_tbl, id, klassp)) return Qtrue;
-
-    /* autoload */
-    if (autoload_tbl && st_lookup(autoload_tbl, id, 0)) {
-	rb_autoload_load(id);
-	*klassp = rb_const_get(rb_cObject, id);
-	return Qtrue;
-    }
-    return Qfalse;
-}
-
-VALUE
-rb_const_get_at(klass, id)
+NORETURN(static void uninitialized_constant _((VALUE, ID)));
+static void
+uninitialized_constant(klass, id)
     VALUE klass;
     ID id;
 {
-    VALUE value;
+    if (klass && klass != rb_cObject)
+	rb_name_error(id, "uninitialized constant %s::%s",
+		      rb_class2name(klass),
+		      rb_id2name(id));
+    else {
+	rb_name_error(id, "uninitialized constant %s", rb_id2name(id));
+    }
+}
 
-    if (RCLASS(klass)->iv_tbl && st_lookup(RCLASS(klass)->iv_tbl, id, &value)) {
-	return value;
-    }
-    if (klass == rb_cObject && top_const_get(id, &value)) {
-	return value;
-    }
-    rb_name_error(id, "uninitialized constant %s::%s",
-		  RSTRING(rb_class_path(klass))->ptr,
-		  rb_id2name(id));
+static VALUE
+const_missing(klass, id)
+    VALUE klass;
+    ID id;
+{
+    return rb_funcall(klass, rb_intern("const_missing"), 1, ID2SYM(id));
+}
+
+
+/*
+ * call-seq:
+ *    mod.const_missing(sym)    => obj
+ *
+ *  Invoked when a reference is made to an undefined constant in
+ *  <i>mod</i>. It is passed a symbol for the undefined constant, and
+ *  returns a value to be used for that constant. The
+ *  following code is a (very bad) example: if reference is made to
+ *  an undefined constant, it attempts to load a file whose name is
+ *  the lowercase version of the constant (thus class <code>Fred</code> is
+ *  assumed to be in file <code>fred.rb</code>). If found, it returns the
+ *  value of the loaded class. It therefore implements a perverse
+ *  kind of autoload facility.
+ *  
+ *    def Object.const_missing(name)
+ *      @looked_for ||= {}
+ *      str_name = name.to_s
+ *      raise "Class not found: #{name}" if @looked_for[str_name]
+ *      @looked_for[str_name] = 1
+ *      file = str_name.downcase
+ *      require file
+ *      klass = const_get(name)
+ *      return klass if klass
+ *      raise "Class not found: #{name}"
+ *    end
+ *  
+ */
+
+VALUE
+rb_mod_const_missing(klass, name)
+    VALUE klass, name;
+{
+    ruby_frame = ruby_frame->prev; /* pop frame for "const_missing" */
+    uninitialized_constant(klass, rb_to_id(name));
     return Qnil;		/* not reached */
 }
 
+static struct st_table *
+check_autoload_table(av)
+    VALUE av;
+{
+    Check_Type(av, T_DATA);
+    if (RDATA(av)->dmark != (RUBY_DATA_FUNC)rb_mark_tbl ||
+	RDATA(av)->dfree != (RUBY_DATA_FUNC)st_free_table) {
+	rb_raise(rb_eTypeError, "wrong autoload table: %s", RSTRING(rb_inspect(av))->ptr);
+    }
+    return (struct st_table *)DATA_PTR(av);
+}
+
 void
-rb_autoload_load(id)
+rb_autoload(mod, id, file)
+    VALUE mod;
+    ID id;
+    const char *file;
+{
+    VALUE av, fn;
+    struct st_table *tbl;
+
+    if (!rb_is_const_id(id)) {
+	rb_raise(rb_eNameError, "autoload must be constant name", rb_id2name(id));
+    }
+    if (!file || !*file) {
+	rb_raise(rb_eArgError, "empty file name");
+    }
+
+    if ((tbl = RCLASS(mod)->iv_tbl) && st_lookup(tbl, id, &av) && av != Qundef)
+	return;
+
+    rb_const_set(mod, id, Qundef);
+    tbl = RCLASS(mod)->iv_tbl;
+    if (st_lookup(tbl, autoload, &av)) {
+	tbl = check_autoload_table(av);
+    }
+    else {
+	av = Data_Wrap_Struct(rb_cData, rb_mark_tbl, st_free_table, 0);
+	st_add_direct(tbl, autoload, av);
+	DATA_PTR(av) = tbl = st_init_numtable();
+    }
+    fn = rb_str_new2(file);
+    FL_UNSET(fn, FL_TAINT);
+    OBJ_FREEZE(fn);
+    st_insert(tbl, id, (st_data_t)rb_node_newnode(NODE_MEMO, fn, ruby_safe_level, 0));
+}
+
+static NODE*
+autoload_delete(mod, id)
+    VALUE mod;
     ID id;
 {
-    char *modname;
-    VALUE module;
+    VALUE val;
+    st_data_t load = 0;
 
-    st_delete(autoload_tbl, &id, (st_data_t *)&modname);
-    if (rb_provided(modname)) {
-	free(modname);
-	return;
+    st_delete(RCLASS(mod)->iv_tbl, (st_data_t*)&id, 0);
+    if (st_lookup(RCLASS(mod)->iv_tbl, autoload, &val)) {
+	struct st_table *tbl = check_autoload_table(val);
+
+	st_delete(tbl, (st_data_t*)&id, &load);
+
+	if (tbl->num_entries == 0) {
+	    DATA_PTR(val) = 0;
+	    st_free_table(tbl);
+	    id = autoload;
+	    if (st_delete(RCLASS(mod)->iv_tbl, (st_data_t*)&id, &val)) {
+		rb_gc_force_recycle(val);
+	    }
+	}
     }
-    module = rb_str_new2(modname);
-    free(modname);
-    FL_UNSET(module, FL_TAINT);
-    rb_f_require(Qnil, module);
+
+    return (NODE *)load;
+}
+
+void
+rb_autoload_load(klass, id)
+    VALUE klass;
+    ID id;
+{
+    VALUE file;
+    NODE *load = autoload_delete(klass, id);
+
+    if (!load || !(file = load->nd_lit) || rb_provided(RSTRING(file)->ptr)) {
+	const_missing(klass, id);
+    }
+    rb_require_safe(file, load->nd_nth);
+}
+
+static VALUE
+autoload_file(mod, id)
+    VALUE mod;
+    ID id;
+{
+    VALUE val, file;
+    struct st_table *tbl;
+    st_data_t load;
+
+    if (!st_lookup(RCLASS(mod)->iv_tbl, autoload, &val) ||
+	!(tbl = check_autoload_table(val)) || !st_lookup(tbl, id, &load)) {
+	return Qnil;
+    }
+    file = ((NODE *)load)->nd_lit;
+    Check_Type(file, T_STRING);
+    if (!RSTRING(file)->ptr) {
+	rb_raise(rb_eArgError, "empty file name");
+    }
+    if (!rb_provided(RSTRING(file)->ptr)) {
+	return file;
+    }
+
+    /* already loaded but not defined */
+    st_delete(tbl, (st_data_t*)&id, 0);
+    if (!tbl->num_entries) {
+	DATA_PTR(val) = 0;
+	st_free_table(tbl);
+	id = autoload;
+	if (st_delete(RCLASS(mod)->iv_tbl, (st_data_t*)&id, &val)) {
+	    rb_gc_force_recycle(val);
+	}
+    }
+    return Qnil;
 }
 
 VALUE
-rb_const_get(klass, id)
+rb_autoload_p(mod, id)
+    VALUE mod;
+    ID id;
+{
+    struct st_table *tbl = RCLASS(mod)->iv_tbl;
+    VALUE val;
+
+    if (!tbl || !st_lookup(tbl, id, &val) || val != Qundef) {
+	return Qnil;
+    }
+    return autoload_file(mod, id);
+}
+
+static VALUE
+rb_const_get_0(klass, id, exclude, recurse)
     VALUE klass;
     ID id;
+    int exclude, recurse;
 {
     VALUE value, tmp;
     int mod_retry = 0;
@@ -1191,30 +1401,62 @@ rb_const_get(klass, id)
     tmp = klass;
   retry:
     while (tmp) {
-	if (RCLASS(tmp)->iv_tbl && st_lookup(RCLASS(tmp)->iv_tbl,id,&value)) {
+	while (RCLASS(tmp)->iv_tbl && st_lookup(RCLASS(tmp)->iv_tbl,id,&value)) {
+	    if (value == Qundef) {
+		rb_autoload_load(tmp, id);
+		continue;
+	    }
+	    if (exclude && tmp == rb_cObject && klass != rb_cObject) {
+		rb_warn("toplevel constant %s referenced by %s::%s",
+			rb_id2name(id), rb_class2name(klass), rb_id2name(id));
+	    }
 	    return value;
 	}
-	if (tmp == rb_cObject && top_const_get(id, &value)) return value;
+	if (!recurse && klass != rb_cObject) break;
 	tmp = RCLASS(tmp)->super;
     }
-    if (!mod_retry && BUILTIN_TYPE(klass) == T_MODULE) {
+    if (!exclude && !mod_retry && BUILTIN_TYPE(klass) == T_MODULE) {
 	mod_retry = 1;
 	tmp = rb_cObject;
 	goto retry;
     }
 
-    /* Uninitialized constant */
-    if (klass && klass != rb_cObject) {
-	rb_name_error(id, "uninitialized constant %s at %s",
-		      rb_id2name(id),
-		      RSTRING(rb_class_path(klass))->ptr);
-    }
-    else { /* global_uninitialized */
-	rb_name_error(id, "uninitialized constant %s",rb_id2name(id));
-    }
-    return Qnil;		/* not reached */
+    return const_missing(klass, id);
 }
 
+VALUE
+rb_const_get_from(klass, id)
+    VALUE klass;
+    ID id;
+{
+    return rb_const_get_0(klass, id, Qtrue, Qtrue);
+}
+
+VALUE
+rb_const_get(klass, id)
+    VALUE klass;
+    ID id;
+{
+    return rb_const_get_0(klass, id, Qfalse, Qtrue);
+}
+
+VALUE
+rb_const_get_at(klass, id)
+    VALUE klass;
+    ID id;
+{
+    return rb_const_get_0(klass, id, Qtrue, Qfalse);
+}
+
+/*
+ *  call-seq:
+ *     remove_const(sym)   => obj
+ *  
+ *  Removes the definition of the given constant, returning that
+ *  constant's value. Predefined classes and singleton objects (such as
+ *  <i>true</i>) cannot be removed.
+ */
+
 VALUE
 rb_mod_remove_const(mod, name)
     VALUE mod, name;
@@ -1223,20 +1465,24 @@ rb_mod_remove_const(mod, name)
     VALUE val;
 
     if (!rb_is_const_id(id)) {
-	rb_name_error(id, "`%s' is not constant", rb_id2name(id));
+	rb_name_error(id, "`%s' is not allowed as a constant name", rb_id2name(id));
     }
     if (!OBJ_TAINTED(mod) && rb_safe_level() >= 4)
 	rb_raise(rb_eSecurityError, "Insecure: can't remove constant");
     if (OBJ_FROZEN(mod)) rb_error_frozen("class/module");
 
-    if (RCLASS(mod)->iv_tbl && st_delete(ROBJECT(mod)->iv_tbl, &id, &val)) {
+    if (RCLASS(mod)->iv_tbl && st_delete(ROBJECT(mod)->iv_tbl, (st_data_t*)&id, &val)) {
+	if (val == Qundef) {
+	    autoload_delete(mod, id);
+	    val = Qnil;
+	}
 	return val;
     }
     if (rb_const_defined_at(mod, id)) {
-	rb_name_error(id, "cannot remove %s::%s", 
+	rb_name_error(id, "cannot remove %s::%s",
 		 rb_class2name(mod), rb_id2name(id));
     }
-    rb_name_error(id, "constant %s::%s not defined", 
+    rb_name_error(id, "constant %s::%s not defined",
 		  rb_class2name(mod), rb_id2name(id));
     return Qnil;		/* not reached */
 }
@@ -1255,18 +1501,6 @@ sv_i(key, value, tbl)
     return ST_CONTINUE;
 }
 
-static int
-autoload_i(key, name, tbl)
-    ID key;
-    const char *name;
-    st_table *tbl;
-{
-    if (!st_lookup(tbl, key, 0)) {
-	st_insert(tbl, key, key);
-    }
-    return ST_CONTINUE;
-}
-
 void*
 rb_mod_const_at(mod, data)
     VALUE mod;
@@ -1279,12 +1513,6 @@ rb_mod_const_at(mod, data)
     if (RCLASS(mod)->iv_tbl) {
 	st_foreach(RCLASS(mod)->iv_tbl, sv_i, (st_data_t)tbl);
     }
-    if ((VALUE)mod == rb_cObject) {
-	st_foreach(rb_class_tbl, sv_i, (st_data_t)tbl);
-	if (autoload_tbl) {
-	    st_foreach(autoload_tbl, autoload_i, (st_data_t)tbl);
-	}
-    }
     return tbl;
 }
 
@@ -1293,10 +1521,12 @@ rb_mod_const_of(mod, data)
     VALUE mod;
     void *data;
 {
+    VALUE tmp = mod;
     for (;;) {
-	data = rb_mod_const_at(mod, data);
-	mod = RCLASS(mod)->super;
-	if (!mod) break;
+	data = rb_mod_const_at(tmp, data);
+	tmp = RCLASS(tmp)->super;
+	if (!tmp) break;
+	if (tmp == rb_cObject && mod != rb_cObject) break;
     }
     return data;
 }
@@ -1325,6 +1555,15 @@ rb_const_list(data)
     return ary;
 }
 
+/*
+ *  call-seq:
+ *     mod.constants    => array
+ *  
+ *  Returns an array of the names of the constants accessible in
+ *  <i>mod</i>. This includes the names of constants in any included
+ *  modules (example at start of section).
+ */
+
 VALUE
 rb_mod_constants(mod)
     VALUE mod;
@@ -1332,27 +1571,40 @@ rb_mod_constants(mod)
     return rb_const_list(rb_mod_const_of(mod, 0));
 }
 
-int
-rb_const_defined_at(klass, id)
+static int
+rb_const_defined_0(klass, id, exclude, recurse)
     VALUE klass;
     ID id;
+    int exclude, recurse;
 {
-    if (RCLASS(klass)->iv_tbl && st_lookup(RCLASS(klass)->iv_tbl, id, 0)) {
-	return Qtrue;
+    VALUE value, tmp;
+    int mod_retry = 0;
+
+    tmp = klass;
+  retry:
+    while (tmp) {
+	if (RCLASS(tmp)->iv_tbl && st_lookup(RCLASS(tmp)->iv_tbl, id, &value)) {
+	    if (value == Qundef && NIL_P(autoload_file(klass, id)))
+		return Qfalse;
+	    return Qtrue;
+	}
+	if (!recurse && klass != rb_cObject) break;
+	tmp = RCLASS(tmp)->super;
     }
-    if (klass == rb_cObject) {
-	return rb_const_defined(klass, id);
+    if (!exclude && !mod_retry && BUILTIN_TYPE(klass) == T_MODULE) {
+	mod_retry = 1;
+	tmp = rb_cObject;
+	goto retry;
     }
     return Qfalse;
 }
 
 int
-rb_autoload_defined(id)
+rb_const_defined_from(klass, id)
+    VALUE klass;
     ID id;
 {
-    if (autoload_tbl && st_lookup(autoload_tbl, id, 0))
-	return Qtrue;
-    return Qfalse;
+    return rb_const_defined_0(klass, id, Qtrue, Qtrue);
 }
 
 int
@@ -1360,20 +1612,15 @@ rb_const_defined(klass, id)
     VALUE klass;
     ID id;
 {
-    VALUE tmp = klass;
+    return rb_const_defined_0(klass, id, Qfalse, Qtrue);
+}
 
-    while (tmp) {
-	if (RCLASS(tmp)->iv_tbl && st_lookup(RCLASS(tmp)->iv_tbl,id,0)) {
-	    return Qtrue;
-	}
-	tmp = RCLASS(tmp)->super;
-    }
-    if (BUILTIN_TYPE(klass) == T_MODULE) {
-	return rb_const_defined(rb_cObject, id);
-    }
-    if (st_lookup(rb_class_tbl, id, 0))
-	return Qtrue;
-    return rb_autoload_defined(id);
+int
+rb_const_defined_at(klass, id)
+    VALUE klass;
+    ID id;
+{
+    return rb_const_defined_0(klass, id, Qtrue, Qfalse);
 }
 
 static void
@@ -1387,14 +1634,25 @@ mod_av_set(klass, id, val, isconst)
 
     if (!OBJ_TAINTED(klass) && rb_safe_level() >= 4)
 	rb_raise(rb_eSecurityError, "Insecure: can't set %s", dest);
-    if (OBJ_FROZEN(klass)) rb_error_frozen("class/module");
+    if (OBJ_FROZEN(klass)) {
+	if (BUILTIN_TYPE(klass) == T_MODULE) {
+	    rb_error_frozen("module");
+	}
+	else {
+	    rb_error_frozen("class");
+	}
+    }
     if (!RCLASS(klass)->iv_tbl) {
 	RCLASS(klass)->iv_tbl = st_init_numtable();
     }
     else if (isconst) {
-	if (st_lookup(RCLASS(klass)->iv_tbl, id, 0) ||
-	    (klass == rb_cObject && st_lookup(rb_class_tbl, id, 0))) {
-	    rb_warn("already initialized %s %s", dest, rb_id2name(id));
+	VALUE value = Qfalse;
+
+	if (st_lookup(RCLASS(klass)->iv_tbl, id, &value)) {
+	    if (value == Qundef)
+		autoload_delete(klass, id);
+	    else
+		rb_warn("already initialized %s %s", dest, rb_id2name(id));
 	}
     }
 
@@ -1410,48 +1668,6 @@ rb_const_set(klass, id, val)
     mod_av_set(klass, id, val, Qtrue);
 }
 
-void
-rb_const_assign(klass, id, val)
-    VALUE klass;
-    ID id;
-    VALUE val;
-{
-    VALUE tmp = klass;
-    
-    while (tmp) {
-	if (RCLASS(tmp)->iv_tbl && st_lookup(RCLASS(tmp)->iv_tbl,id,0)) {
-	    st_insert(RCLASS(tmp)->iv_tbl, id, val);
-	    return;
-	}
-	tmp = RCLASS(tmp)->super;
-    }
-    /* pre-defined class */
-    if (st_lookup(rb_class_tbl, id, 0)) {
-	st_delete(rb_class_tbl, &id, 0);
-	st_insert(RCLASS(rb_cObject)->iv_tbl, id, val);
-	return;
-    }
-
-    /* autoload */
-    if (autoload_tbl && st_lookup(autoload_tbl, id, 0)) {
-	char *modname;
-
-	st_delete(autoload_tbl, (st_data_t *)&id, (st_data_t *)&modname);
-	free(modname);
-	st_insert(RCLASS(rb_cObject)->iv_tbl, id, val);
-	return;
-    }
-
-    /* Uninitialized constant */
-    if (klass && klass != rb_cObject)
-	rb_name_error(id, "uninitialized constant %s::%s",
-		      RSTRING(rb_class_path(klass))->ptr,
-		      rb_id2name(id));
-    else {
-	rb_name_error(id, "uninitialized constant %s",rb_id2name(id));
-    }
-}
-
 void
 rb_define_const(klass, name, val)
     VALUE klass;
@@ -1460,6 +1676,9 @@ rb_define_const(klass, name, val)
 {
     ID id = rb_intern(name);
 
+    if (!rb_is_const_id(id)) {
+	rb_warn("rb_define_const: invalide name `%s' for constant", name);
+    }
     if (klass == rb_cObject) {
 	rb_secure(4);
     }
@@ -1631,6 +1850,23 @@ cv_i(key, value, ary)
     return ST_CONTINUE;
 }
 
+/*
+ *  call-seq:
+ *     mod.class_variables   => array
+ *  
+ *  Returns an array of the names of class variables in <i>mod</i> and
+ *  the ancestors of <i>mod</i>.
+ *     
+ *     class One
+ *       @@var1 = 1
+ *     end
+ *     class Two < One
+ *       @@var2 = 2
+ *     end
+ *     One.class_variables   #=> ["@@var1"]
+ *     Two.class_variables   #=> ["@@var2", "@@var1"]
+ */
+
 VALUE
 rb_mod_class_variables(obj)
     VALUE obj;
@@ -1647,6 +1883,26 @@ rb_mod_class_variables(obj)
     return ary;
 }
 
+/*
+ *  call-seq:
+ *     remove_class_variable(sym)    => obj
+ *  
+ *  Removes the definition of the <i>sym</i>, returning that
+ *  constant's value.
+ *     
+ *     class Dummy
+ *       @@var = 99
+ *       puts @@var
+ *       remove_class_variable(:@@var)
+ *       puts(defined? @@var)
+ *     end
+ *     
+ *  <em>produces:</em>
+ *     
+ *     99
+ *     nil
+ */
+
 VALUE
 rb_mod_remove_cvar(mod, name)
     VALUE mod, name;
@@ -1661,11 +1917,11 @@ rb_mod_remove_cvar(mod, name)
 	rb_raise(rb_eSecurityError, "Insecure: can't remove class variable");
     if (OBJ_FROZEN(mod)) rb_error_frozen("class/module");
 
-    if (RCLASS(mod)->iv_tbl && st_delete(ROBJECT(mod)->iv_tbl, &id, &val)) {
+    if (RCLASS(mod)->iv_tbl && st_delete(ROBJECT(mod)->iv_tbl, (st_data_t*)&id, &val)) {
 	return val;
     }
     if (rb_cvar_defined(mod, id)) {
-	rb_name_error(id, "cannot remove %s for %s", 
+	rb_name_error(id, "cannot remove %s for %s",
 		 rb_id2name(id), rb_class2name(mod));
     }
     rb_name_error(id, "class variable %s not defined for %s",
